/*  FixParser.cpp : Defines the entry point for the DLL application. */
 
/*
 * Licensed Materials - Property of IBM
 * 5648-C63
 * (C) Copyright IBM Corp. 1999
 */

/******************************************************************************/
/*                                                                            */
/* Note: This sample plugin node implementation is not necessarily portable   */
/* to all platforms supported by MQSeries Integrator Version 2.0.1            */
/*                                                                            */
/******************************************************************************/

/* Includes */

#ifdef WIN32
#include <wchar.h>
#include <windows.h>
#include "largeint.h"
#endif

#include <malloc.h>
#include <memory.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#if defined _AIX
#include <sys/time.h>
#endif

#include <BipCpi.h>

#include "common.h"
#include "fixparser.h"
#include "fixsub.h"
#include "trace.h"
#include "metadata.h"
#include "mgmtmsg.h"
#include "parsubs.h"

/**************************************************************/
/*                                                            */
/* Defines and global variables                               */
/*                                                            */
/**************************************************************/

static char copyright[] = "(C) Copyright IBM Corp, 2000";
static char Version[]=\
"@(#)FixParser V1.00 - MQSI V2.01 Fix Message Parser  - Jim MacNair ";

#ifdef _DEBUG
static char Level[]="FixParser V1.0 Debug version ("__DATE__" "__TIME__")";
#else
static char Level[]="FixParser V1.0 Release version ("__DATE__" "__TIME__")";
#endif

	int		parserStartHigh=0;	// Time parser was started (high word)
	int		parserStartLow=0;	// Time parser was started (low word)
#ifdef WIN32
	__int64 totalParseTime=0;	// Total parsing time in microseconds
	__int64 totalWriteTime=0;	// Total time to build a buffer in microseconds
	__int64 maxParseTime=0;		// Maximum time to parse a message in microseconds
	__int64 minParseTime=-1;	// Minimum time to parse a message in microseconds
	__int64 maxWriteTime=0;		// Maximum time to build a buffer in microseconds
	__int64 minWriteTime=-1;	// Minimum time to build a buffer in microseconds
#else
	long long totalParseTime=0;	// Total parsing time in microseconds
	long long totalWriteTime=0;	// Total time to build a buffer in microseconds
	long long maxParseTime=0;	// Maximum time to parse a message in microseconds
	long long minParseTime=-1;	// Minimum time to parse a message in microseconds
	long long maxWriteTime=0;	// Maximum time to build a buffer in microseconds
	long long minWriteTime=-1;	// Minimum time to build a buffer in microseconds
#endif
	int     maxMessage=0;		// Largest message processed
	int     minMessage=-1;		// Smallest message processed
	int     totalBytes=0;		// Number of bytes in all messages processed
	int		totalWritten=0;		// Number of bytes written
	int		maxWriteLength=0;	// Largest message written
	int		minWriteLength=-1;	// Smallest message written
	int     messageCount=0;		// Number of input messages processed    
	int     writeCount=0;		// Number of cpiWriteBuffer calls processed

	static int contextCount=0;	// Number of parser objects 

#ifdef _DEBUG
extern int debug;
extern int traceParser;
extern int traceWrite;
extern int traceModule;
extern int traceMgmt;
#endif

extern int	inTagCount;			// Number of tags parsed in input messages

void saveMsgStats(PARSER_CONTEXT_ST* context)

{
#ifdef _DEBUG
	if (traceModule)
	{
		Log("\nPLUGIN: -> saveMsgStats() context=0x%x totalParseTime=%d iSize=%d iMsgType=%d\n", 
			context, context->totalParseTime, context->iSize, context->iMsgType);
	}
#endif

	/* check for a data vs a management message */
	/* if the previous message caused an exception, then don't gather the times */
	/* since they may be lower than what it takes to parse a message */
	if ((0 == context->iMsgType) && (0 == context->iBadMsg))
	{
		/* data message, save the statistics for this message */
		if ((int) (context->totalParseTime) > maxParseTime)
		{
			/* get the time in microseconds */
			maxParseTime = (int) (context->totalParseTime); 
		}

		if (-1 == minParseTime)
		{
			/* get the time in milliseconds */
			minParseTime = (int) (context->totalParseTime);
		}

		if ((int) (context->totalParseTime) < minParseTime)
		{
			/* get the time in milliseconds */
			minParseTime = (int) (context->totalParseTime);
		}
	}

#ifdef _DEBUG
	if (traceModule)
	{
		Log("\nPLUGIN: <- saveMsgStats() maxParseTime=%d, minParseTime=%d, maxMessage=%d, minMessage=%d\n", 
			maxParseTime, minParseTime, maxMessage, minMessage);
	}
#endif
}

/******************************************************************************/
/*                                                                            */
/* Routine to initialize the context area.                                    */
/*                                                                            */
/******************************************************************************/

void resetContext(PARSER_CONTEXT_ST* context)

{
	if (context->iBuffer != NULL)
	{
		free(context->iBuffer);
		context->iBuffer = NULL;
	}

	// initialize the context area to binary zeros
	memset(context, 0, sizeof(PARSER_CONTEXT_ST));
}

/******************************************************************************/
/*                                                                            */
/* Routine to capture the starting time and save it in the context.           */
/*                                                                            */
/******************************************************************************/

void getStartTime(PARSER_CONTEXT_ST* context)

{
#ifdef WIN32
	int				rc;
	LARGE_INTEGER	start;
#elif defined _AIX
	timebasestruct_t start;
#endif

#ifdef WIN32
	// Get the current timer value
	rc = QueryPerformanceCounter(&start);
	if (0 == rc)
	{
		/* no high-res timer available */
		context->iStartTime = 0;
	}
	else
	{
		/* Save the start time in the context */
		context->iStartTime = start.QuadPart;
	}
#elif defined _AIX
	read_real_time(&start, TIMEBASE_SZ);
	time_base_to_time(&start, TIMEBASE_SZ);
	context->iStartTime.tb_high = start.tb_high;
	context->iStartTime.tb_low  = start.tb_low;
#else
	context->iStartTime = 0;
#endif
}

/******************************************************************************/
/*                                                                            */
/* Routine to calculate the difference between the current time and the       */
/* start time in the context and return the difference as a 64-bit            */
/* integer (expressed in microseconds).                                       */
/*                                                                            */
/******************************************************************************/

#ifdef WIN32
__int64 calculateTime(PARSER_CONTEXT_ST* context)

{
	int				rc;
	__int64			time=0;
	LARGE_INTEGER	tempLarge;
	__int64			timerFreq;
	__int64			endTime=0;

#ifdef _DEBUG
	if (traceModule)
	{
		Log("\nPLUGIN: -> calculateTime() iStartTimeHigh=%I64d\n", 
			context->iStartTime);
	}
#endif

	/* Get the current timer value */
	rc = QueryPerformanceCounter(&tempLarge);
	if (0 != rc)
	{
		/* save the ending count value */
		endTime = tempLarge.QuadPart;

		rc = QueryPerformanceFrequency(&tempLarge);
		if (0 != rc)
		{
			/* get the frequency value */
			timerFreq = tempLarge.QuadPart;

			/* calculate the elapsed time in microsecs */
			time = ((endTime - context->iStartTime) * 1000000) / timerFreq;
		}
	}

#ifdef _DEBUG
	if (traceModule)
	{
		Log("context->iStartTime=%I64d, endTime=%I64d\n", context->iStartTime, endTime);
		Log("\nPLUGIN: <- calculateTime() time=%d\n", time);
	}
#endif

	return time;
}
#else
long long calculateTime(PARSER_CONTEXT_ST* context)

{
	int				rc;
	long long		time=0;
	timebasestruct_t endTime;
	int				seconds;
	int				microsecs;

#ifdef _DEBUG
	if (traceModule)
	{
		Log("\nPLUGIN: -> calculateTime() iStartTime=(%d,%d)\n", 
			context->iStartTime.tb_high, context->iStartTime.tb_low);
	}
#endif

	// get the current time in seconds and nano-seconds
	read_real_time(&endTime, TIMEBASE_SZ);
	time_base_to_time(&endTime, TIMEBASE_SZ);

	// get the seconds and nanoseconds
	seconds = endTime.tb_high - context->iStartTime.tb_high;
	microsecs = (endTime.tb_low - context->iStartTime.tb_low + 500) / 1000;

	// check if the end nanoseconds was less than the start nanoseconds
	if (microsecs < 0)
	{
		// decrease the seconds and make sure the nanoseconds are positive
		seconds--;
		microsecs += 1000000;
	}

	// calculate the total time in microseconds
	time = (seconds * 1000000) + microsecs;

#ifdef _DEBUG
	if (traceModule)
	{
		Log(" endTime.HighPart=%d, endTime.LowPart=%d\n", endTime.tb_high, endTime.tb_low);
		Log(" seconds=%d, microsecs=%d\n", seconds, microsecs);
		Log("\nPLUGIN: <- calculateTime() time=%lld in microseconds\n", time);
	}
#endif

	return time;
}
#endif

/******************************************************************************/
/*                                                                            */
/* Plugin Parser Implementation Function:                                     */
/*                                                                            */
/******************************************************************************/

CciContext* cpiCreateContext(CciParser* parser)

{
  PARSER_CONTEXT_ST *pc;

#ifdef _DEBUG
	if (traceModule)
	{
		Log("PLUGIN: -> cpiCreateContext() parser=0x%x\n", parser);
	}
#endif

	/* Allocate a pointer to the local context */
	pc = (PARSER_CONTEXT_ST*)malloc(sizeof(PARSER_CONTEXT_ST));
	memset(pc, 0, sizeof(PARSER_CONTEXT_ST));
	pc->iBuffer = NULL;

	if (pc) 
	{
		/* Clear the context area */
		resetContext(pc);

		/* Keep count of the number of parser objects */
		contextCount++;

		/* initialize the message type to -1 so we */
		/* don't collect statistics before the first message */
		pc->iMsgType = -1;
	}
	else 
	{
#ifdef _DEBUG
		/* Error: Unable to allocate memory */
		if (traceModule)
		{
			Log("Unable to allocate memory for Context for Fix parser\n");
		}
#endif
	}

 #ifdef _DEBUG
	if (traceModule)
	{
		Log("PLUGIN: <- cpiCreateContext(0x%x) count %d\n", pc, contextCount);
	}
#endif

	/* Return the pointer to the local context */
	return((CciContext*) pc);
}

/******************************************************************************/
/*                                                                            */
/* Plugin Parser Implementation Function:                                     */
/*                                                                            */
/******************************************************************************/

int cpiParseBuffer(CciParser*  parser, CciContext* context, int offset)

{
	PARSER_CONTEXT_ST*	pc = (PARSER_CONTEXT_ST*)context;
	int					rc;
	int					size;

#ifdef WIN32
	__int64				time;
#ifdef _DEBUG
	SYSTEMTIME			curtime;
#endif
#elif defined _AIX
	long long			time;
#ifdef _DEBUG
	timebasestruct_t	curtime;
#endif
#endif

#ifdef _DEBUG
#ifdef WIN32
	if (traceModule) 
	{
		// Get the current time
		GetSystemTime(&curtime);

		Log("\nPLUGIN: -> cpiParseBuffer() %d/%d/%d %2.2d:%2.2d:%2.2d.%d \n",
			curtime.wYear,
			curtime.wMonth,
			curtime.wDay,
			curtime.wHour,
			curtime.wMinute,
			curtime.wSecond,
			curtime.wMilliseconds);
	}
#elif defined _AIX
	if (traceModule) 
	{
		// get the current time
		read_real_time(&curtime, TIMEBASE_SZ);

		Log("\nPLUGIN: -> cpiParseBuffer()\n");
	}
#else
	if (traceModule) 
	{
		Log("\nPLUGIN: -> cpiParseBuffer()\n");
	}
#endif

	if (traceModule) 
	{
		Log(" parser (0x%x) context (0x%x) offset (%d)\n", 
			parser, context, offset);
	}
#endif

	/* check if we have statistics to save */
	if (pc->iSize > 0)
	{
		saveMsgStats(pc);
	}

	/* initialize the message context */
	resetContext(pc);

	/* Start the timer */
	getStartTime(pc);

	/* Save size of the buffer */
	pc->iSize = cpiBufferSize(&rc, parser) - offset;

	/* get the code page of the input data */
	pc->iCodePage = getCodePage(parser);

	/* make sure we have a sensible default code page */
	if (pc->iCodePage == 0)
	{
		pc->iCodePage = DEF_CODEPAGE;
	}

	/* get the numeric encoding for the message */
	pc->iEncoding = getEncoding(parser);

	/* make sure we have a proper encoding value */
	if (pc->iEncoding == 0)
	{
		/* set a sensible default */
		pc->iEncoding = DEF_ENCODING;
	}

	/* Get a pointer to the message buffer and save the current offset */
	pc->iInput = (unsigned char *)cpiBufferPointer(&rc, parser);

	if (pc->iInput != NULL) 
	{
		pc->iBuffer = (CciChar *)malloc((pc->iSize + 1) * sizeof(CciChar));
		memset(pc->iBuffer, 0, (pc->iSize + 1) * sizeof(CciChar));

		/* translate the input data to Unicode */
		/* this translation may not translate the entire message, */
		/* since raw data elements may contain a binary zero in   */
		/* the middle of the message, and this will cause the     */
		/* translation function to stop before it has translated  */
		/* the entire message.  However, this initial translation */
		/* will the beginning of the message, up until the point  */
		/* where a raw data field is found.  When a raw data      */
		/* field is processed, the rest of the message will be    */
		/* translated again, to make sure the entire message is   */
		/* translated before it is parsed.                        */
		size = cciMbsToUcs(&rc, (char *) pc->iInput + offset, pc->iBuffer, pc->iSize, pc->iCodePage);
		pc->iBuffer[pc->iSize] = 0;
	}
	else 
	{
		pc->iIndex = 0;
		pc->iBuffer = NULL;
	}

	/* initialize the index within the buffer */
	pc->iIndex = 0;

	/* Prime the first byte in the stream */
	size = cciUcsToMbs(&rc, 
					   pc->iBuffer + pc->iIndex, 
					   &pc->iCurrentCharacter, 
					   1, 
					   DEF_CODEPAGE);

	/* Set the current element to the body element */
	pc->eBody = cpiRootElement(&rc, parser);
	pc->eCurrentElement = pc->eBody;

	/* get the domain, message set and type values */
	getMessageDomain(parser, pc->msgDomain, sizeof(pc->msgDomain) - 1);
	getMessageSet(parser, pc->msgSet, sizeof(pc->msgSet) - 1);
	getMessageType(parser, pc->msgType, sizeof(pc->msgType) - 1);

	/* check if we should use encryption/decryption */
	if (strcmp(pc->msgDomain, NO_ENCRYPT_DOMAIN) != 0)
	{
		pc->iUseEncryption = 1;
	}

	/* claim the whole management message */
	size = pc->iSize;

	/* check for a systems management message instead of a normal message */
	if (strcmp(pc->msgSet, MGMT_MSG) == 0)
	{
		/* remember this is a systems management message */
		pc->iMsgType = 1;
	}
	else
	{
		/* normal FIX format */
		/* Check if this is a valid Fix message */
		checkMessageFormat(pc);

		/* Increment the message count */
		messageCount++;

		/* get the total number of bytes processed */
		totalBytes += pc->iSize;

		if (pc->iSize > maxMessage)
		{
			maxMessage = pc->iSize;
		}

		if (-1 == minMessage)
		{
			minMessage = pc->iSize;
		}

		if (pc->iSize < minMessage)
		{
			minMessage = pc->iSize;
		}

		// at this point, we will drop the check sum
		// by subtracting 7 bytes from the total message length
		pc->iSize -= 7;
		inTagCount++;

		/* add to the total time */
		time = calculateTime(pc);
		totalParseTime += time;
		pc->totalParseTime += time;
	}

#ifdef _DEBUG
	if (traceModule) 
	{
		Log("PLUGIN: <- cpiParseBuffer() buffer 0x%x message size =%d\n", 
			pc->iBuffer, pc->iSize);
	}
#endif

	/* We will assume ownership of the remainder of the buffer */
	return(size - offset);
}

/******************************************************************************/
/*                                                                            */
/* Plugin Parser Implementation Function:                                     */
/*                                                                            */
/******************************************************************************/

void cpiParseFirstChild(CciParser*  parser, CciContext* context, CciElement* element)

{
	PARSER_CONTEXT_ST*	pc = (PARSER_CONTEXT_ST*) context;
	int					rc;

#ifdef WIN32
	__int64				time;
#else
	long long			time;
#endif

#ifdef _DEBUG
	CciChar	name[256];
	CciSize	length;
	CciBool	completePrev;
	CciBool completeNext;
#endif

#ifdef _DEBUG
	if (traceModule) 
	{
		memset(name, 0, sizeof(name));
		length = cpiElementName(&rc, element, name, sizeof(name));
		completePrev = cpiElementCompletePrevious(&rc, element);
		completeNext = cpiElementCompleteNext(&rc, element);
		Log("PLUGIN: -> cpiParseFirstChild() context=0x%x element %S (0x%x) bits (%d %d)\n",
			context, name, element, completePrev, completeNext);
	}
#endif

	// Start the timer
	getStartTime(pc);

	// Check if we have a systems management message or a normal one
	if (0 == pc->iMsgType)
	{
		while ((cpiElementCompletePrevious(&rc, element) != COMPLETE_BIT) &&
			   (pc->eCurrentElement != CCI_NULL_ADDR))
		{
			parseNextItem(parser, pc);

#ifdef _DEBUG
			if (traceModule) 
			{
				length = cpiElementName(&rc, element, name, sizeof(name));
				completePrev = cpiElementCompletePrevious(&rc, element);
				completeNext = cpiElementCompleteNext(&rc, element);
				Log(" returned from parseNextItem element=%S (0x%x) bits (%d %d)\n",
					name, element, completePrev, completeNext);
			}
#endif
		}

		/* add to the total time */
		time = calculateTime(pc);
		totalParseTime += time;
		pc->totalParseTime += time;
	}
	else
	{
		// system management message
		// handle as special case
		// This will build a complete parse tree
		parseMgmtMsg(parser, pc);
	}

#ifdef _DEBUG
	if (traceModule) 
	{
		completePrev = cpiElementCompletePrevious(&rc, element);
		completeNext = cpiElementCompleteNext(&rc, element);
		Log("PLUGIN: <- cpiParseFirstChild() element %S (0x%x) bits (%d %d)\n", 
			name, element, completePrev, completeNext);
	}
#endif

  return;
}

/******************************************************************************/
/*                                                                            */
/* Plugin Parser Implementation Function:                                     */
/*                                                                            */
/******************************************************************************/
void cpiParseLastChild(CciParser*  parser,
					   CciContext* context,
					   CciElement* element)

{
	PARSER_CONTEXT_ST*	pc = (PARSER_CONTEXT_ST*)context;
	int					rc;

#ifdef WIN32
	__int64				time;
#else
	long long			time;
#endif

#ifdef _DEBUG
	CciChar	name[256];
	CciSize	length;
	CciBool	completePrev;
	CciBool completeNext;
#endif

#ifdef _DEBUG
	if (traceModule) 
	{
		memset(name, 0, sizeof(name));
		length = cpiElementName(&rc, element, name, sizeof(name));
		completePrev = cpiElementCompletePrevious(&rc, element);
		completeNext = cpiElementCompleteNext(&rc, element);
		Log("PLUGIN: -> cpiParseLastChild() %S (0x%x) bits(%d %d)\n",
			name, element, completePrev, completeNext);
	}
#endif

	// Start the timer
	getStartTime(pc);

	// Check if we have a systems management message or a normal one
	if (0 == pc->iMsgType)
	{
		while ((cpiElementCompleteNext(&rc, element) != COMPLETE_BIT) &&
			   (pc->eCurrentElement != CCI_NULL_ADDR))
		{
			parseNextItem(parser, pc);
		}

		/* add to the total time */
		time = calculateTime(pc);
		totalParseTime += time;
		pc->totalParseTime += time;
	}
	else
	{
		// system management message
		// handle as special case
		// This will build a complete parse tree
		parseMgmtMsg(parser, pc);
	}

#ifdef _DEBUG
	if (traceModule) 
	{
		completePrev = cpiElementCompletePrevious(&rc, element);
		completeNext = cpiElementCompleteNext(&rc, element);
		Log("PLUGIN: <- cpiParseLastChild() %S (0x%x) bits(%d %d)\n",
			name, element, completePrev, completeNext);
	}
#endif

	return;
}

/******************************************************************************/
/*                                                                            */
/* Plugin Parser Implementation Function:                                     */
/*                                                                            */
/******************************************************************************/
void cpiParseLeftSibling(CciParser*  parser,
						 CciContext* context,
						 CciElement* element)

{
	PARSER_CONTEXT_ST*	pc = (PARSER_CONTEXT_ST*)context;
	int					rc;

#ifdef WIN32
	__int64				time;
#else
	long long			time;
#endif

#ifdef _DEBUG
	CciChar	name[256];
	CciSize	length;
	CciBool	completePrev;
	CciBool completeNext;
#endif

#ifdef _DEBUG
	if (traceModule) 
	{
		memset(name, 0, sizeof(name));
		length = cpiElementName(&rc, element, name, sizeof(name));
		completePrev = cpiElementCompletePrevious(&rc, element);
		completeNext = cpiElementCompleteNext(&rc, element);
		Log("PLUGIN: -> cpiParseLeftSibling() %S (0x%x) bits(%d %d)\n",
			name, element, completePrev, completeNext);
	}
#endif

	// Start the timer
	getStartTime(pc);

	// Check if we have a systems management message or a normal one
	if (0 == pc->iMsgType)
	{
		// Normal message, parse the next group until the element is completely parsed
		while ((cpiElementCompletePrevious(&rc, cpiParent(&rc, element)) != COMPLETE_BIT) &&
			   (cpiPreviousSibling(&rc, element) == CCI_NULL_ADDR) &&
			   (pc->eCurrentElement != CCI_NULL_ADDR))
		{
			parseNextItem(parser, pc);
		}

		/* add to the total time */
		time = calculateTime(pc);
		totalParseTime += time;
		pc->totalParseTime += time;
	}
	else
	{
		// system management message
		// handle as special case
		// This will build a complete parse tree
		parseMgmtMsg(parser, pc);
	}

#ifdef _DEBUG
	if (traceParser) 
	{
		completePrev = cpiElementCompletePrevious(&rc, element);
		completeNext = cpiElementCompleteNext(&rc, element);
		Log("PLUGIN: <- cpiParseLeftSibling() bits(%d %d)\n", completePrev, completeNext);
	}
#endif

	return;
}

/******************************************************************************/
/*                                                                            */
/* Plugin Parser Implementation Function:                                     */
/*                                                                            */
/******************************************************************************/
void cpiParseRightSibling(CciParser*  parser,
						  CciContext* context,
						  CciElement* element)

{
	PARSER_CONTEXT_ST*	pc = (PARSER_CONTEXT_ST*)context;
	int					rc;

#ifdef WIN32
	__int64				time;
#else
	long long			time;
#endif

#ifdef _DEBUG
	CciChar	name[256];
	CciSize	length;
	CciBool	completePrev;
	CciBool completeNext;
#endif

#ifdef _DEBUG
	if (traceModule > 0) 
	{
		memset(name, 0, sizeof(name));
		length = cpiElementName(&rc, element, name, sizeof(name));
		completePrev = cpiElementCompletePrevious(&rc, element);
		completeNext = cpiElementCompleteNext(&rc, element);
		Log("PLUGIN: -> cpiParseRightSibling() context=0x%x element=0x%x %S bits(%d %d)\n",
			context, element, name, completePrev, completeNext);
	}
#endif

	// Start the timer
	getStartTime(pc);

	// Check if we have a systems management message or a normal one
	if (0 == pc->iMsgType)
	{
		while ((cpiElementCompleteNext(&rc, cpiParent(&rc, element)) != COMPLETE_BIT) &&
			   (cpiNextSibling(&rc, element) == CCI_NULL_ADDR) &&
			   (pc->eCurrentElement != CCI_NULL_ADDR))
		{
			parseNextItem(parser, pc);
		}

		/* add to the total time */
		time = calculateTime(pc);
		totalParseTime += time;
		pc->totalParseTime += time;
	}
	else
	{
		// system management message
		// handle as special case
		// This will build a complete parse tree
		parseMgmtMsg(parser, pc);
	}

#ifdef _DEBUG
   if (traceModule > 0) 
   {
		completePrev = cpiElementCompletePrevious(&rc, element);
		completeNext = cpiElementCompleteNext(&rc, element);
		Log("PLUGIN: <- cpiParseRightSibling() bits(%d %d)\n", completePrev, completeNext);
   }
#endif

	return;
}

/******************************************************************************/
/*                                                                            */
/* Plugin Parser Implementation Function:                                     */
/*                                                                            */
/******************************************************************************/
int cpiWriteBuffer(CciParser*  parser,
				   CciContext*	context)

{
	PARSER_CONTEXT_ST*	pc = (PARSER_CONTEXT_ST*)context;
	int					initialSize = 0;
	int					rc = 0;
	int					length;
	int					msecs;

#ifdef WIN32
	__int64				time;
#else
	long long			time;
#endif

#ifdef _DEBUG
	if (traceModule) 
	{
		Log("PLUGIN: -> cpiWriteBuffer() parser=0x%x context=0x%x\n", parser, context);
	}
#endif

	/* Reset the context area */
	resetContext(pc);

	// Start the timer
	getStartTime(pc);

	/* get the code page from the properties */
	pc->iCodePage = getCodePage(parser);

	/* make sure we have a sensible default code page */
	if (pc->iCodePage == 0)
	{
		pc->iCodePage = DEF_CODEPAGE;
	}

	/* get the numeric encoding from the properties for the message */
	pc->iEncoding = getEncoding(parser);

	/* make sure we have a proper encoding value */
	if (pc->iEncoding == 0)
	{
		/* set a sensible default */		
		pc->iEncoding = DEF_ENCODING;
	}

	/* get the domain, message set and type values */
	getMessageDomain(parser, pc->msgDomain, sizeof(pc->msgDomain) - 1);
	getMessageSet(parser, pc->msgSet, sizeof(pc->msgSet) - 1);
	getMessageType(parser, pc->msgType, sizeof(pc->msgType) - 1);

	/* check if we should use encryption/decryption */
	if (strcmp(pc->msgDomain, NO_ENCRYPT_DOMAIN) != 0)
	{
		pc->iUseEncryption = 1;
	}

	/* get the body element as the current element */
	pc->eBody = cpiRootElement(&rc, parser);

	if (CCI_SUCCESS != rc)
	{
		/* get the body element as the last child of the root */
		pc->eCurrentElement = CCI_NULL_ADDR;
	}

	/* create the output buffer from the logical message model */
	length = createOutputMessage(parser, pc);

	/* make sure we were able to build the output message */
	/* before we gather statistics */
	if ((0 == pc->iBadMsg) && (length > 0))
	{
		// Keep a count of the output messages
		writeCount++;

		// keep track of the total bytes written
		totalWritten += length;

		// Calculate the elapsed time in microseconds
		time = calculateTime(pc);
		msecs = (int) (time);

		if (msecs > maxWriteTime)
		{
			maxWriteTime = msecs;
		}

		/* capture this amount if it is the first one */
		if (-1 == minWriteTime)
		{
			minWriteTime = msecs;
		}

		if (msecs < minWriteTime)
		{
			minWriteTime = msecs;
		}

		if (length > maxWriteLength)
		{
			maxWriteLength = length;
		}

		/* check for first time through */
		if (-1 == minWriteLength)
		{
			minWriteLength = length;
		}

		if (length < minWriteLength)
		{
			minWriteLength = length;
		}

		// remember the time spent
		totalWriteTime += time;
	}

#ifdef _DEBUG
	if (traceModule) 
	{
		Log("PLUGIN: <- cpiWriteBuffer() length=%d, time=%d\n", length, time);
	}
#endif

	return cpiBufferSize(0, parser) - initialSize;
}

/******************************************************************************/
/*                                                                            */
/* Plugin Parser Implementation Function:                                     */
/*                                                                            */
/******************************************************************************/
void cpiDeleteContext(CciParser*  parser,
					  CciContext* context)

{
	PARSER_CONTEXT_ST*	pc = (PARSER_CONTEXT_ST*)context;
	int					rc = 0;
	RPT_ITEM			*rptPtr;
	RPT_ITEM			*nextrpt;

#ifdef _DEBUG
	if (traceModule) 
	{
		Log("PLUGIN: -> cpiDeleteContext() parser=0x%x context=0x%x\n",
			parser, context);
	}
#endif

	/* free the input buffer if present */
	if (pc->iBuffer != NULL)
	{
		free(pc->iBuffer);
		pc->iBuffer = NULL;
	}

	/* free the output buffer if present */
	if (pc->iOutput != NULL)
	{
		free(pc->iOutput);
		pc->iOutput = NULL;
	}

	/* check for a repeating data item chain */
	rptPtr = pc->iRepeat;
	while (rptPtr != NULL)
	{
		nextrpt = (RPT_ITEM *)rptPtr->nextItem;
		free(rptPtr);
		rptPtr = nextrpt;
	}

	// Free the context area
	free(pc);

#ifdef _DEBUG
	if (traceParser) 
	{
		Log("PLUGIN: <- cpiDeleteContext()\n");
	}
#endif

	return;
}

/******************************************************************************/
/*                                                                            */
/* Plugin Parser Implementation Function:                                     */
/*                                                                            */
/******************************************************************************/

CciBool cpiParserType(CciParser*  parser, CciContext* context)

{
	/* we are a body parser */
	return TRUE;
}

/******************************************************************************/
/*                                                                            */
/* Plugin Parser Implementation Function:                                     */
/*                                                                            */
/******************************************************************************/
#ifdef WIN32
void __declspec(dllexport) *bipGetParserFactory()
#else
void *bipGetParserFactory()
#endif

{
	/* Declare variables */
	CciFactory*     factoryObject;
	int             rc;
	static CPI_VFT  vftable;
#ifdef WIN32
  	SYSTEMTIME		curtime;
	FILETIME		filtime;
#elif defined _AIX
	timebasestruct_t time;
#endif

#ifdef WIN32
	// Get the current time
	GetSystemTime(&curtime);
	SystemTimeToFileTime(&curtime, &filtime);

	// Save the time the parser was started
	parserStartHigh = filtime.dwHighDateTime;
	parserStartLow  = filtime.dwLowDateTime;
#elif defined _AIX
	// get the current time in seconds and nanoseconds
	read_real_time(&time, TIMEBASE_SZ);
	time_base_to_time(&time, TIMEBASE_SZ);

	// save the time the parser was started
#else
	parserStartHigh = 0;
	parserStartLow = 0;
#endif

	/* Setup function table with pointers to parser implementation functions */
	memset(&vftable, 0, sizeof(vftable));
	vftable.iFpCreateContext            = cpiCreateContext;
	vftable.iFpParseBuffer              = cpiParseBuffer;
	vftable.iFpParseFirstChild          = cpiParseFirstChild;
	vftable.iFpParseLastChild           = cpiParseLastChild;
	vftable.iFpParseLeftSibling         = cpiParseLeftSibling;
	vftable.iFpParseRightSibling        = cpiParseRightSibling;
	vftable.iFpWriteBuffer              = cpiWriteBuffer;
	vftable.iFpDeleteContext            = cpiDeleteContext;
	vftable.iFpIsHeaderParser           = cpiParserType;
	vftable.iFpSetValue                 = NULL;
	vftable.iFpValue                    = NULL;
	vftable.iFpNextParserClassName      = NULL;
	vftable.iFpSetNextParserClassName   = NULL;
	vftable.iFpNextParserEncoding       = NULL;
	vftable.iFpNextParserCodedCharSetId = NULL;

#ifdef _DEBUG
	/* check for DEBUG requested */
	initializeTrace();
#endif

	/* Initialize the file table for the meta data cache in memory */
	initializeMetadata();

	/* Create the parser factory for this plugin */
	factoryObject = cpiCreateParserFactory(&rc, PARSER_NAME);
	if (factoryObject) {
		/* Define the classes of message supported by the factory */
		cpiDefineParserClass(&rc, factoryObject, PARSER_CLASS,  &vftable);
		cpiDefineParserClass(&rc, factoryObject, PARSER_CLASS2,  &vftable);
	}
	else 
	{
		/* Error: Unable to create parser factory */ 
#ifdef _DEBUG
		if (traceModule) 
		{
			Log("Unable to create parser factory, rc=%d\n", rc);
		}
#endif
	}

	/* Return address of this factory object to the broker */
	return(factoryObject);
}

const char * getFixParserLevel()

{
	return Level;
}