/*  fixsub.cpp  */

/*
 * Licensed Materials - Property of IBM
 * 5648-C63
 * (C) Copyright IBM Corp. 2000
 */

#ifdef WIN32
#include <wchar.h>
#include <windows.h>
#endif

#include <malloc.h>
#include <memory.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <BipCci.h>
#include <BipCni.h>


#include "common.h"
#include "fixsub.h"
#include "parsubs.h"
#include "trace.h"
#include "metadata.h"

#include "fixerr.h"

#include "BipSampPluginUtil.h"

/* following added to fix a bug in the BipCci.h file on AIX */
#ifndef WIN32
typedef struct CciByteArray CciByteArray;
#endif

/* maximum number of characters for a type field */
#define MAX_TYPE_LEN	16

#ifdef _DEBUG
static char Level[]="Fixsub.cpp V1.0 Debug Version ("__DATE__" "__TIME__")";
#else
static char Level[]="Fixsub.cpp V1.0 Release Version("__DATE__" "__TIME__")";
#endif

	int     badMessageCount=0;		// Number of messages rejected
	int		badOutputCount=0;		// Number of output messages rejected
	int     inTagCount=0;			// Number of tags processed on input
	int     outTagCount=0;			// Number of tags processed on input

extern FILETBL *fileTable;			// pointer to metadata for text items

#ifdef _DEBUG
extern int debug;
extern int traceParser;
extern int traceWrite;
extern int traceModule;
extern int traceMgmt;
#endif

/******************************************************************************/
/*                                                                            */
/* Plugin Parser Utility Function: Skip white space from current position     */
/*                                                                            */
/******************************************************************************/

void skipWhiteSpace(PARSER_CONTEXT_ST* context)

{
	CciChar * p = context->iBuffer + context->iIndex; 

	/* Advance until end of white-space or end of buffer */
	while ((context->iIndex != context->iSize) && (CciCharIsSpace(*p)))
	{
		advance(context);
		if (context->iIndex == context->iSize) return;
	}

	return;
}

/******************************************************************************/
/*                                                                            */
/* Plugin Parser Utility Function: Advance to next character                  */
/*                                                                            */
/******************************************************************************/

void advance(PARSER_CONTEXT_ST* context)

{
	int rc;
	int size;

#ifdef _DEBUG
	if (traceModule) 
	{
		Log("PLUGIN: -> advance(): index=%d char=%C\n", 
			context->iIndex, context->iCurrentCharacter);
	}
#endif

	/* Don't run past the end of the buffer */
	if (context->iIndex == context->iSize) return;

	/* Advance to the next character */
	context->iIndex++;

	/* Detect and handle the end condition */
	if (context->iIndex == context->iSize) return;

	/* Obtain the next character from the buffer           */
	/* as a native character in context->iCurrentCharacter */
	size = cciUcsToMbs(&rc, 
					   context->iBuffer + context->iIndex, 
					   &context->iCurrentCharacter, 
					   1, 
					   DEF_CODEPAGE);

#ifdef _DEBUG
	if (traceModule) 
	{
		Log("PLUGIN: <- advance(): index=%d char=%C\n",
			context->iIndex, context->iCurrentCharacter);
	}
#endif
}

/******************************************************************************/
/*                                                                            */
/* This routine should be replaced with whatever encryption algorithm is      */
/* to be used.  The dataIn parameter points to the data to be encrypted, and  */
/* the dataLen parameter gives the length in bytes.  The dataOut parameter    */
/* points to a storage buffer to hold the encrypted data and the routine      */
/* should return the length of the encrypted data as an integer.  There is    */
/* an assumption that the encrypted data will not be substantially larger     */
/* then the original data.  If this is a possibility, then the length of      */
/* allocated storage area may need to be increased.  A small (<128 bytes)     */
/* fixed increase in size should not present a problem, unless multiple       */
/* encrypted areas are used within a single message (not likely).             */
/*                                                                            */
/* A very simple obfuscation routine is provided in the encrypt and decrypt   */
/* routines, to provide a simple sample of how encrypted messages would be    */
/* handled.  These routines should NOT be used in a production application.   */
/* The routines are commented out and have been replaced with a single line   */
/* that merely copies the input to the output.  If encryption is to be used,  */
/* the encrypt and decrypt routines must be adapted to use the real           */
/* encryption software and the indicated line removed.                        */
/*                                                                            */
/******************************************************************************/

int encrypt(const unsigned char * dataIn, const int dataLen, unsigned char * dataOut)

{
	int				i=0;
#ifdef _ENCRYPT_TEST
	int				count=0;
	unsigned char	newch; 
#endif

	while (i < dataLen)
	{
#ifdef _ENCRYPT_TEST
		switch (count)
		{
		case 0:
			{
				dataOut[i] = dataIn[i] ^ 143;
				break;
			}
		case 1:
			{
				newch = (dataIn[i] >> 1) + ((dataIn[i] & 1) << 7);
				dataOut[i] = newch ^ 156;
				break;
			}
		case 2:
			{
				newch = (dataIn[i] >> 2) + ((dataIn[i] & 3) << 6);
				dataOut[i] = newch ^ 212;
				break;
			}
		case 3:
			{
				newch = (dataIn[i] >> 3) + ((dataIn[i] & 7) << 5);
				dataOut[i] = newch ^ 43;
				break;
			}
		case 4:
			{
				newch = (dataIn[i] >> 4) + ((dataIn[i] & 15) << 4);
				dataOut[i] = newch ^ 75;
				break;
			}
		case 5:
			{
				newch = (dataIn[i] >> 5) + ((dataIn[i] & 31) << 3);
				dataOut[i] = newch ^ 255;
				break;
			}
		case 6:
			{
				newch = (dataIn[i] >> 6) + ((dataIn[i] & 63) << 2);
				dataOut[i] = newch ^ 15;
				break;
			}
		case 7:
			{
				newch = (dataIn[i] >> 7) + ((dataIn[i] & 127) << 1);
				dataOut[i] = newch ^ 195;
				break;
			}
		}

		count++;
		count = count % 8;
#else
		dataOut[i] = dataIn[i];
#endif
		i++;
	}

	return i;
}

int decrypt(const unsigned char * dataIn, const int dataLen, unsigned char * dataOut)

{
	int				i=0;
#ifdef _ENCRYPT_TEST
	int				count=0;
	unsigned char	newch; 
#endif

	while (i < dataLen)
	{
#ifdef _ENCRYPT_TEST
		switch (count)
		{
		case 0:
			{
				dataOut[i] = dataIn[i] ^ 143;
				break;
			}
		case 1:
			{
				newch = dataIn[i] ^ 156;
				dataOut[i] = ((newch & 127) << 1) + ((newch & 128) >> 7);
				break;
			}
		case 2:
			{
				newch = dataIn[i] ^ 212;
				dataOut[i] = ((newch & 63) << 2) + ((newch & 192) >> 6);
				break;
			}
		case 3:
			{
				newch = dataIn[i] ^ 43;
				dataOut[i] = ((newch & 31) << 3) + ((newch & 224) >> 5);
				break;
			}
		case 4:
			{
				newch = dataIn[i] ^ 75;
				dataOut[i] = ((newch & 15) << 4) + ((newch & 240) >> 4);
				break;
			}
		case 5:
			{
				newch = dataIn[i] ^ 255;
				dataOut[i] = ((newch & 7) << 5) + ((newch & 248) >> 3);
				break;
			}
		case 6:
			{
				newch = dataIn[i] ^ 15;
				dataOut[i] = ((newch & 3) << 6) + ((newch & 252) >> 2);
				break;
			}
		case 7:
			{
				newch = dataIn[i] ^ 195;
				dataOut[i] = ((newch & 1) << 7) + ((newch & 254) >> 1);
				break;
			}
		}

		count++;
		count = count % 8;
#else
		dataOut[i] = dataIn[i];
#endif
		i++;
	}

	return i;
}

/******************************************************************************/
/*                                                                            */
/* Routine to calculate a FIX check sum value.                                */
/*                                                                            */
/******************************************************************************/

int calcCheckSum(unsigned char *data, const int length)

{
	int		sum=0;
	int		i=0;

#ifdef _DEBUG
	if (traceModule) 
	{
		Log("PLUGIN: -> calcCheckSum(): length=%d\n", length);
	}
#endif

	while (i < length)
	{
		sum += data[i++];
	}

	sum = sum % 256;

#ifdef _DEBUG
	if (traceModule) 
	{
		Log("PLUGIN: <- calcCheckSum(): sum=%d\n", sum);
	}
#endif

	return sum;
}

/******************************************************************************/
/*                                                                            */
/* Routine to verify that the check sum value in the message matches the      */
/* check sum value calculated from the message itself.                        */
/*                                                                            */
/******************************************************************************/

void validateCheckSum(PARSER_CONTEXT_ST* context)

{
	int			rc;
	int			checksum1=0;
	int			checksum2=0;
	CciChar		*Uofs;
	char		workArea[32];
	char		checksumchar1[32];
	char		checksumchar2[32];
	char		errmsg[512];

#ifdef _DEBUG
	if (traceModule) 
	{
		Log("PLUGIN: -> validateCheckSum(): length=%d\n", context->iSize - 7);
	}
#endif

	/* first, check if there is a CheckSum at the end of the message */
	/* we have to translate the data into Unicode first to ensure    */
	/* that we can work with any input message                       */
	Uofs = context->iBuffer + context->iSize - 7;
	cciMbsToUcs(&rc,
				(char *) context->iInput + context->iSize - 7,
				Uofs,
				7,
				context->iCodePage);

	/* translate back into a work area, dropping the trailing delimiter */
	cciUcsToMbs(&rc,
				Uofs,
				workArea,
				6,
				DEF_CODEPAGE);

	if (memcmp(workArea, CHECK_SUM_TAG, sizeof(CHECK_SUM_TAG) - 1) != 0)
	{
		/* missing check sum tag */
		/* Increase the invalid message counter */
		badMessageCount++;
		context->iBadMsg = 1;

		/* get the message text */
		strcpy(errmsg, "No CheckSum item at end of message");

#ifdef _DEBUG
		if (traceModule) 
		{
			/* get the error message as a string */
			Log("%s\n", errmsg);
		}
#endif
		/* throw an exception */
		cciThrowException(&rc, 
						  CCI_PARSER_EXCEPTION, 
						  __FILE__, 
						  __LINE__, 
						  "validateCheckSum",
						  ERRDICT,
						  NO_CHECKSUM,
						  errmsg,
						  0);
	}

	checksum2 = atoi(workArea + sizeof(CHECK_SUM_TAG) - 1);

	/* calculate the check sum from the data */
	checksum1 = calcCheckSum(context->iInput, context->iSize - 7);

	if (checksum1 != checksum2)
	{
		/* check sums don't match - throw an exception */
		/* Increase the invalid message counter */
		badMessageCount++;
		context->iBadMsg = 1;

		sprintf(errmsg, "Checksum tag value (%d) does not match calculated sum (%d)", 
				checksum2, checksum1);

#ifdef _DEBUG
		if (traceModule) 
		{
			/* get the error message as a string */
			Log("%s\n", errmsg);
		}
#endif
		/* throw an exception */
		sprintf(checksumchar1, "%d", checksum1);
		sprintf(checksumchar2, "%d", checksum2);
		cciThrowException(&rc, 
						  CCI_PARSER_EXCEPTION, 
						  __FILE__, 
						  __LINE__, 
						  "validateCheckSum",
						  ERRDICT,
						  BAD_CHECKSUM,
						  errmsg,
						  checksumchar1,
						  checksumchar2,
						  0);
	}

#ifdef _DEBUG
	if (traceModule) 
	{
		Log("PLUGIN: <- validateCheckSum(): checksum=%d\n", checksum2);
	}
#endif
}

/******************************************************************************/
/*                                                                            */
/* Routine to insert an equal sign into the output stream.                    */
/*                                                                            */
/******************************************************************************/

void insertEqual(PARSER_CONTEXT_ST* context)

{
	int		rc;
	int		size;

	/* insert the equal sign in the desired code page */
	size = cciUcsToMbs(&rc,
					   STR_EQUAL_SIGN,
					   (char *) context->iOutput + context->iIndex,
					   context->iRemaining,
					   context->iCodePage);

	/* account for the character that we added */
	context->iIndex++;
	context->iRemaining--;
}

/******************************************************************************/
/*                                                                            */
/* Plugin Parser Utility Function: Advance to end of a tag                    */
/*                                                                            */
/******************************************************************************/

void goToDelimiter(PARSER_CONTEXT_ST* context, char delimiter)

{
#ifdef _DEBUG
	if (traceModule) 
	{
		Log("PLUGIN: -> goToDelimiter(): index=%d\n", context->iIndex);
	}
#endif

	/* Advance until a non-name character is found */
	while ((context->iCurrentCharacter != delimiter) &&
		   (context->iIndex != context->iSize))
	{
		advance(context);
	}

#ifdef _DEBUG
	if (traceModule) 
	{
		Log("PLUGIN: <- goToDelimiter(): index=%d\n", context->iIndex);
	}
#endif
}

/******************************************************************************/
/*                                                                            */
/* Routine to isolate a tag into a string area.                               */
/*                                                                            */
/******************************************************************************/

int isolateTag(PARSER_CONTEXT_ST* context)

{
	int			rc;
	int			startMarker;
	int			markedSize;
	char		errmsg[512];

#ifdef _DEBUG
	if (traceModule) 
	{
		Log("PLUGIN: -> isolateTag(): index=%d\n", context->iIndex);
	}
#endif

	/* Get a pointer to the start of the tag */
	startMarker = context->iIndex;

	/* Find end of tag */
//	goToDelimiter(context, chEqual);

	/* Advance until a non-name character is found */
	while ((context->iCurrentCharacter != chEqual) &&
		   (context->iIndex != context->iSize))
	{
		/* check if the delimiter is too long */
		if (context->iIndex - startMarker > MAX_TAG_LEN)
		{
			/* tag is too long to be valid */
			/* Increase the invalid message counter */
			badMessageCount++;
			context->iBadMsg = 1;

			sprintf(errmsg, "tag too long at offset %d\n", context->iIndex);

#ifdef _DEBUG
			if (traceModule) 
			{
				/* get the error message as a string */
				Log("%s\n", errmsg);
			}
#endif
			/* throw an exception */
			cciThrowException(&rc, 
							  CCI_PARSER_EXCEPTION, 
							  __FILE__, 
							  __LINE__, 
							  "isolateTag",
							  ERRDICT,
							  TAG_TOO_LONG,
							  errmsg,
							  0);
		}
	
		/* check for an invalid character */
		if ((context->iCurrentCharacter < '0') ||
			(context->iCurrentCharacter > '9'))
		{
			/* tag contains an invalid character */
			/* Increase the invalid message counter */
			badMessageCount++;
			context->iBadMsg = 1;

			sprintf(errmsg, "Invalid character in tag at offset %d\n", context->iIndex);

#ifdef _DEBUG
			if (traceModule) 
			{
				/* get the error message as a string */
				Log("%s\n", errmsg);
			}
#endif
			/* throw an exception */
			cciThrowException(&rc, 
							  CCI_PARSER_EXCEPTION, 
							  __FILE__, 
							  __LINE__, 
							  "isolateTag",
							  ERRDICT,
							  INVALID_TAG_CHAR,
							  errmsg,
							  0);
		}

		/* move on to the next character */
		advance(context);
	}


	/* Compute the size of the tag */
	markedSize = context->iIndex - startMarker;

	/* delimit the string by changing the equal */
	/* sign to a null value.                    */
	context->iBuffer[context->iIndex] = 0;

	/* skip the delimiter */
	advance(context);

#ifdef _DEBUG
	if (traceModule) 
	{
		Log("PLUGIN: <- isolateTag(): tag(%S) length=%d, new index=%d\n", 
			context->iBuffer + startMarker, markedSize, context->iIndex);
	}
#endif

	return markedSize;
}

/******************************************************************************/
/*                                                                            */
/* Routine to isolate the data for a particular element.                      */
/*                                                                            */
/******************************************************************************/

int isolateData(PARSER_CONTEXT_ST* context)

{
	int			startMarker;
	int			markedSize;

#ifdef _DEBUG
	if (traceModule) 
	{
		Log("PLUGIN: -> isolateData(): index=%d\n", context->iIndex);
	}
#endif

	/* Get a pointer to the start of the tag */
	startMarker = context->iIndex;

	// Find end of tag
	goToDelimiter(context, DELIMITER);

	/* Compute the size of the tag */
	markedSize = context->iIndex - startMarker;

	/* delimit the data as a string by changing */
	/* the delimiter to a null value.           */
	context->iBuffer[context->iIndex] = 0;

	/* skip the delimiter */
	advance(context);

#ifdef _DEBUG
	if (traceModule) 
	{
		Log("PLUGIN: <- isolateData(): data(%S) length=%d, new index=%d\n", 
			context->iBuffer + startMarker, markedSize, context->iIndex);
	}
#endif

	return markedSize;
}

/**************************************************************/
/*                                                            */
/* Routine to validate if this is a valid Fix message.        */
/*                                                            */
/* To be considered a valid message, the first data segment   */
/* must be a type "8" segment.                                */
/*                                                            */
/**************************************************************/

void checkMessageFormat(PARSER_CONTEXT_ST* context)

{
	int			rc;
	int			i;
	int			size;
	char		errmsg[512];
	CciChar		*ptr;
	CciChar		messageType[MAX_FIX];

#ifdef _DEBUG
	if (traceModule) 
	{
		Log("PLUGIN: -> checkMessageFormat() Beginning of Message(%8.8S) codepage(%d)\n", 
			context->iBuffer, context->iCodePage);
	}
#endif

	if (memcmp(context->iBuffer, BEGIN_STRING, sizeof(BEGIN_STRING) - sizeof(CciChar)) != 0)
	{
		// Increase the invalid message counter
		badMessageCount++;
		context->iBadMsg = 1;

		// get the error message as a string
		strcpy(errmsg, "No begin string element found at front of message");

#ifdef _DEBUG
		if (traceModule) 
		{
			Log("%s\n", errmsg);
		}
#endif
		// We don't have a message header group
		cciThrowException(&rc, 
						  CCI_PARSER_EXCEPTION, 
						  __FILE__, 
						  __LINE__, 
						  "checkMessageFormat",
						  ERRDICT,
						  NO_BEGINSTRING,
						  errmsg,
						  0);
	}

	// Next, check if we can find the end of the begin string
	// First, find the SOH for the basic header
	ptr = context->iBuffer + CciCharLen(BEGIN_STRING);
	i=0;
	while ((ptr[0] != DELIMITER) && 
		   (ptr < context->iBuffer + context->iSize) &&
		   (i < sizeof(messageType) - 1))
	{
		if (ptr[0] != chPeriod)
		{
			messageType[i++] = ptr++[0];
		}
		else
		{
			ptr++;
		}
	}

	/* terminate the string properly */
	messageType[i] = 0;

	/* translate to local code page that we are running on */
	size = cciUcsToMbs(&rc, 
					   messageType, 
					   context->sLevelName, 
					   sizeof(context->sLevelName), 
					   DEF_CODEPAGE);

	// Did we find the end of the basic header?
	if (ptr[0] != DELIMITER)
	{
		/* No, this message is invalid - raise an error */
		/* Increase the invalid message counter */
		badMessageCount++;
		context->iBadMsg = 1;

		/* get the error message as a string */
		strcpy(errmsg, "No delimiter (0x01) found for BeginString data");

#ifdef _DEBUG
		if (traceModule) 
		{
			Log("%s\n", errmsg);
		}
#endif
		/* We don't the BeginString at the beginning of the message */
		cciThrowException(&rc, 
						  CCI_PARSER_EXCEPTION, 
						  __FILE__, 
						  __LINE__, 
						  "checkMessageFormat",
						  ERRDICT,
						  NO_BEGINSTRING_DELIMITER,
						  errmsg,
						  0);
	}

	/* check if we have a delimiter at the end of the message */
	if (DELIMITER != context->iInput[context->iSize - 1])
	{
		/* required delimiter (0x01) not found at end of message */
		/* No, this message is invalid - raise an error */
		// Increase the invalid message counter
		badMessageCount++;
		context->iBadMsg = 1;

		// get the error message as a string
		strcpy(errmsg, "No ending delimiter character (0x01) found at end of message");

#ifdef _DEBUG
		if (traceModule) 
		{
			Log("%s\n", errmsg);
		}
#endif
		// We don't have a message header group
		cciThrowException(&rc, 
						  CCI_PARSER_EXCEPTION, 
						  __FILE__, 
						  __LINE__, 
						  "checkMessageFormat",
						  ERRDICT,
						  NO_ENDING_DELIMITER,
						  errmsg,
						  0);
	}

	/* finally, verify that the message matches the checksum */
	validateCheckSum(context);

#ifdef _DEBUG
	if (traceModule) 
	{
		Log("PLUGIN: <- checkMessageFormat()\n");
	}
#endif
}

/******************************************************************************/
/*                                                                            */
/* Plugin Parser Utility Function: Sets level complete and makes the parent   */
/* the current element.                                                       */
/*                                                                            */
/******************************************************************************/

void finishHeaderElement(PARSER_CONTEXT_ST* context)

{
	int		rc;
#ifdef _DEBUG
	CciSize size;
	CciChar name[512];
	CciBool	completePrev;
	CciBool completeNext;
#endif

#ifdef _DEBUG
	if (traceModule) 
	{
		completePrev = cpiElementCompletePrevious(&rc, context->eCurrentElement);
		completeNext = cpiElementCompleteNext(&rc, context->eCurrentElement);
		size = cpiElementName(&rc, context->eCurrentElement, name, sizeof(name) - 1); 
		Log("PLUGIN: -> finishHeaderElement(); currentElement=%S (0x%x) bits(%d %d)\n", 
			name, context->eCurrentElement, completePrev, completeNext);
	}
#endif

	// set the last child complete bit for the current element
	cpiSetElementCompleteNext(&rc, context->eCurrentElement, COMPLETE_BIT);

	// restore the main element as the current parent
	context->eCurrentElement = cpiParent(&rc, context->eCurrentElement);

#ifdef _DEBUG
	if (traceModule) 
	{
		completePrev = cpiElementCompletePrevious(&rc, context->eCurrentElement);
		completeNext = cpiElementCompleteNext(&rc, context->eCurrentElement);
		size = cpiElementName(&rc, context->eCurrentElement, name, sizeof(name) - 1); 
		Log("PLUGIN: <- finishHeaderElement(); currentElement=%S (0x%x) bits(%d %d)\n",
			name, context->eCurrentElement, completePrev, completeNext);
	}
#endif

	return;
}

/******************************************************************************/
/*                                                                            */
/* Plugin Parser Utility Function: Creates Specified name tag.                */
/*                                                                            */
/******************************************************************************/

CciElement * createHeaderElement(CciParser* parser,
								 PARSER_CONTEXT_ST* context,
								 CciChar* tagName)

{
	CciElement*        element;
	int                rc;
#ifdef _DEBUG
	CciSize size;
	CciChar name[512];
	CciBool	completePrev;
	CciBool completeNext;
#endif

#ifdef _DEBUG
	if (traceModule) 
	{
		name[0] = 0;
		size = cpiElementName(&rc, context->eCurrentElement, name, sizeof(name) - 1); 
		Log("PLUGIN: -> createHeaderElement(); tagName=%S parent=%S (0x%x)\n", 
			tagName, name, context->eCurrentElement);
	}
#endif

	/* Create a new name element for the tag */
	element = cpiCreateAndInitializeElement(&rc, 
					parser,
					CCI_ELEMENT_TYPE_NAME,
					tagName,
					COMPLETE_BIT,
					0);

	/* Add the element as the last child of the current parent */
	cpiAddAsLastChild(&rc, context->eCurrentElement, element);

	// mark the first child pointer of the parent as complete
	cpiSetElementCompletePrevious(&rc, context->eCurrentElement, COMPLETE_BIT);

	// make the element the new element the current parent
	context->eCurrentElement=cpiLastChild(&rc, context->eCurrentElement);

#ifdef _DEBUG
	if (traceModule) 
	{
		completePrev = cpiElementCompletePrevious(&rc, element);
		completeNext = cpiElementCompleteNext(&rc, element);
		Log("PLUGIN: <- createHeaderElement(); created element=%S (0x%x) bits (%d %d)\n", 
			tagName, element, completePrev, completeNext);
	}
#endif

	return element;
}

/******************************************************************************/
/*                                                                            */
/* Plugin Parser Utility Function: Creates Specified TAG with fixed length    */
/*                                 value                                      */
/*                                                                            */
/******************************************************************************/

void createFixedLengthElement(CciParser* parser,
							  PARSER_CONTEXT_ST* context,
							  CciChar* tagName,
							  int length)

{
	CciElement*	element;
	int			rc;
	int			size;
#ifdef _DEBUG
	int			len;
	CciChar		data[128];
#endif

#ifdef _DEBUG
	if (traceModule) 
	{
		Log("PLUGIN: -> createFixedLengthElement(); tagName=%S, length=%d\n",
            tagName, length);
	}
#endif

	/* Create and set a new value element for the text value */
	element = cpiCreateAndInitializeElement(&rc, 
				parser,
				CCI_ELEMENT_TYPE_NAME_VALUE,
				tagName,
				COMPLETE_BIT,
				COMPLETE_BIT);

	// set the element value
	cpiSetCharacterValueFromBuffer(&rc, element, context->iBuffer + context->iIndex, length);

	// insert the new element into the parse tree
	cpiAddAsLastChild(&rc, context->eCurrentElement, element);

	/*Move index to end of length requested */
	context->iIndex += length;

	/* get the new current character, since we jumped in the buffer */
	size = cciUcsToMbs(&rc, 
					   context->iBuffer + context->iIndex, 
					   &context->iCurrentCharacter, 
					   1, 
					   DEF_CODEPAGE);

#ifdef _DEBUG
	if (traceModule) 
	{
		len = length;
		if (length > sizeof(data) - 1)
		{
			len = sizeof(data) - 1;
		}

		memcpy(data, context->iBuffer + context->iIndex, len * sizeof(CciChar));
		data[len] = 0;
		Log(" added element %S (0x%x) length=%d offset now=%d data (%S)\n", 
			tagName, element, length, data, context->iIndex);

		Log("PLUGIN: <- createFixedLengthElement(); element=%S (0x%x) value=%S\n", 
			tagName, element, data);
	}
#endif

	return;
}


/******************************************************************************/
/*                                                                            */
/* Plugin Parser Utility Function: Creates Specified TAG with variable length */
/*                                 value                                      */
/*                                                                            */
/******************************************************************************/

void createVariableLengthElement(CciParser* parser,
								 PARSER_CONTEXT_ST* context,
								 CciChar* tagName,
								 CciChar* data,
								 int length)

{
	CciElement*		element;
	int				rc;

#ifdef _DEBUG
	if (traceModule) 
	{
		Log("PLUGIN: -> createVariableLengthElement(); tagName=%S\n", tagName);
	}
#endif

	/* Create a new name element for the tag */
	element = cpiCreateAndInitializeElement(&rc, 
				parser,
				CCI_ELEMENT_TYPE_NAME_VALUE,
				tagName,
				COMPLETE_BIT,
				COMPLETE_BIT);

	// Set the element value
	cpiSetCharacterValueFromBuffer(&rc, element, data, length); 

	// add this element as the last child
	cpiAddAsLastChild(&rc, context->eCurrentElement, element);

#ifdef _DEBUG
	if (traceModule) 
	{
		Log("PLUGIN: <- createVariableLengthElement(); last element=0x%x\n", element);
	}
#endif

	return;
}

int checkIfMember(PARSER_CONTEXT_ST* context, CciChar* typePtr)

{
	int			result = -1;
	CciChar		*nextone;
	RPT_ITEM	*rptPtr;

	if (NULL == context->iRepeat)
	{
		return result;
	}

	rptPtr = context->iRepeat;
	nextone = rptPtr->members;

	/* check for another first item */
	if (CciCharCmp(rptPtr->typePtr, typePtr) == 0)
	{
		result = 0;
	}
	else
	{
		while ((nextone[0] != 0) && (CciCharCmp(nextone, typePtr) != 0))
		{
			nextone += CciCharLen(nextone) + 1;
		}

		if (nextone[0] != 0)
		{
			result = 0;
		}
	}

	return result;
}

/******************************************************************************/
/*                                                                            */
/*  Routine to check if a variable is a the count field for a repeating data  */
/*  item.  The message type captured from the MsgType tag (#35) must match    */
/*  the table and the tag name must also match an entry in the repeating      */
/*  elements table.                                                           */
/*                                                                            */
/******************************************************************************/

int checkIfRpt(PARSER_CONTEXT_ST* context, 
			   FILETBL* fileTable,
			   CciChar* typePtr,
			   int varnum)

{
	int		result=-1;
	int		i=0;

	/* search for a match for this element type and message type */
	while ((i < fileTable->maxrpt) && 
		   ((strcmp(fileTable->rvar[i].Name, context->sMsgType) != 0) ||
		   (CciCharCmp(context->iPrevTag, fileTable->UnamePtr + fileTable->rvar[i].CountUtypeptr) != 0) ||
		   (CciCharCmp(typePtr, fileTable->UnamePtr + fileTable->rvar[i].FirstUtypeptr) != 0)))
	{
		i++;
	}

	if (i < fileTable->maxrpt)
	{
		result = i;
	}

	return result;
}

/******************************************************************************/
/*                                                                            */
/* Routine to handle raw data elements.  Raw data elements can contain any    */
/* character in the data area.  The previous element is supposed to contain   */
/* the length of the raw data.  This length is used to identify where the     */
/* raw data ends.  The raw data will be stored as a byte array.  No           */
/* translation or other processing of the data will be done.                  */
/*                                                                            */
/******************************************************************************/

void processRawData(CciParser* parser, 
					PARSER_CONTEXT_ST* context, 
					CciChar *namePtr,
					unsigned char flags,
					FILETBL* fileTable)

{
	int				rc;
	int				rawLength=0;
	int				saveSize;
	int				saveIndex;
	int				newSize;
	CciElement*		element;
	CciByteArray	tempArray;
	unsigned char	*tempBuf;
	unsigned char	*saveInp;
	CciChar			*UtempBuf;
	CciChar			*saveBuf;
	char			tempLen[32];
	char			charlen[32];
	char			chartype[32];
	char			charemain[32];
	char			errmsg[512];

#ifdef _DEBUG
	if (traceModule) 
	{
		Log("PLUGIN: -> processRawData(): name=%S\n", namePtr);
	}
#endif

	/* determine the length of the raw data    */
	/* the previous field should be an integer */
	/* but we will accept a string as well     */
	if ((context->iPrevDataType != DATA_INTEGER) &&
		(context->iPrevDataType != DATA_STRING))
	{
		/* The previous fields is not of the right type */
		/* throw an exception */
		/* Increase the invalid message counter */
		badMessageCount++;
		context->iBadMsg = 1;

		/* get the error message as a string */
		sprintf(errmsg, "Invalid data type (%d) in length field for raw data item", context->iPrevDataType);

#ifdef _DEBUG
		if (traceModule) 
		{
			Log("%s\n", errmsg);
		}
#endif
		/* Write the error to the event viewer and terminate */
		sprintf(chartype, "%d", context->iPrevData);
		cciThrowException(&rc, 
						  CCI_PARSER_EXCEPTION, 
						  __FILE__, 
						  __LINE__, 
						  "processRawData",
						  ERRDICT,
						  BAD_LENGTH_FIELD_TYPE,
						  errmsg,
						  chartype,
						  0);
	}

	/* get the previous data into the local code page */
	cciUcsToMbs(&rc, context->iPrevData, tempLen, sizeof(tempLen) - 1, DEF_CODEPAGE);

	/* convert to an integer */
	rawLength = atoi(tempLen);

	if (0 >= rawLength)
	{
		/* length field evaluates to zero */
		/* throw an exception             */
		/* Increase the invalid message counter */
		badMessageCount++;
		context->iBadMsg = 1;

		/* get the error message as a string */
		sprintf(errmsg, "length field for raw data at offset (%d) field evaluates to zero or negative number", context->iIndex);

#ifdef _DEBUG
		if (traceModule) 
		{
			Log("%s\n", errmsg);
		}
#endif
		/* Write the error to the event viewer and terminate */
		sprintf(chartype, "%d", context->iIndex);
		cciThrowException(&rc, 
						  CCI_PARSER_EXCEPTION, 
						  __FILE__, 
						  __LINE__, 
						  "processRawData",
						  ERRDICT,
						  ZERO_LENGTH_FIELD,
						  errmsg,
						  chartype,
						  0);
	}

	if (context->iSize < context->iIndex + rawLength)
	{
		/* length is too large  - past the end of the buffer */
		/* throw an exception */
		/* Increase the invalid message counter */
		badMessageCount++;
		context->iBadMsg = 1;

		/* get the error message as a string */
		sprintf(errmsg, "length field (%d) for raw data at offset (%d) field exceeds remaining buffer\n", 
				rawLength, context->iIndex, context->iSize - context->iIndex);

#ifdef _DEBUG
		if (traceModule) 
		{
			Log("%s\n", errmsg);
		}
#endif
		/* Write the error to the event viewer and terminate */
		sprintf(charlen, "%d", rawLength);
		sprintf(chartype, "%d", context->iIndex);
		sprintf(charemain, "%d", context->iSize - context->iIndex);
		cciThrowException(&rc, 
						  CCI_PARSER_EXCEPTION, 
						  __FILE__, 
						  __LINE__, 
						  "processRawData",
						  ERRDICT,
						  LENGTH_FIELD_TOO_LARGE,
						  errmsg,
						  charlen,
						  chartype,
						  charemain,
						  0);
	}

	/* check if the character after the data is a delimiter */
	if (context->iInput[context->iIndex + rawLength] != DELIMITER)
	{
		/* delimiter is missing - throw an exception */
		/* Increase the invalid message counter */
		badMessageCount++;
		context->iBadMsg = 1;

		/* get the error message as a string */
		sprintf(errmsg, "No delimiter at end of raw data at offset(%d) - length (%d)", 
			context->iIndex, rawLength);

#ifdef _DEBUG
		if (traceModule) 
		{
			Log("%s\n", errmsg);
		}
#endif
		/* Write the error to the event viewer and terminate */
		sprintf(chartype, "%d", context->iIndex);
		sprintf(charlen, "%d", rawLength);
		cciThrowException(&rc, 
						  CCI_PARSER_EXCEPTION, 
						  __FILE__, 
						  __LINE__, 
						  "processRawData",
						  ERRDICT,
						  MISSING_DELIMITER_RAW_DATA,
						  errmsg,
						  chartype,
						  charlen,
						  0);
	}

	if ((flags & FLAGS_DECRYPT_PARSE) == FLAGS_DECRYPT_PARSE)
	{
		/* we need to decrypt this item */
		/* save the current size, index and buffer */
		saveSize = context->iSize;
		saveBuf	= context->iBuffer;
		saveInp = context->iInput;
		saveIndex = context->iIndex;

		/* allocate a temporary area to hold the decrypted data */
		tempBuf = (unsigned char *)malloc(rawLength + 256);
		UtempBuf = (CciChar *)malloc((rawLength + 256) * 2);

		/* decrypt the data */
		newSize = decrypt(context->iInput + context->iIndex, rawLength, tempBuf);

		/* translate the decrypted area to Unicode */
		cciMbsToUcs(&rc,
					(char *)tempBuf,
					UtempBuf,
					newSize,
					context->iCodePage);

		/* prepare to parse the new input area */
		context->iInput = tempBuf;
		context->iBuffer = UtempBuf;
		context->iIndex = 0;
		context->iSize = newSize;

		/* parse the new area */
		while (context->iIndex < context->iSize)
		{
			parseElement(parser, context, fileTable);
		}

		/* restore the context area */
		context->iInput = saveInp;
		context->iBuffer = saveBuf;
		context->iSize = saveSize;
		context->iIndex = saveIndex + rawLength;

		/* free the storage areas */
		free(tempBuf);
		free(UtempBuf);
	}
	else
	{
		/* first, add the value to the current element */
		/* Create a new name element for the tag */
		element = cpiCreateAndInitializeElement(&rc, 
					parser,
					CCI_ELEMENT_TYPE_NAME_VALUE,
					namePtr,
					COMPLETE_BIT,
					COMPLETE_BIT);

		if ((flags & FLAGS_DECRYPT_ONLY) == FLAGS_DECRYPT_ONLY)
		{
			/* allocate a temporary area to hold the decrypted data */
			tempBuf = (unsigned char *)malloc(rawLength + 256);
			UtempBuf = (CciChar *)malloc((rawLength + 256) * 2);

			/* decrypt the data and then translate it to Unicode */
			newSize = decrypt(context->iInput + context->iIndex, rawLength, tempBuf);
			cciMbsToUcs(&rc,
						(char *)tempBuf,
						UtempBuf,
						newSize,
						context->iCodePage);

			/* add the data to the element */
			cpiSetCharacterValueFromBuffer(&rc, element, UtempBuf, newSize); 

			/* free the storage areas */
			free(tempBuf);
			free(UtempBuf);
		}
		else
		{
			/* we have the length and the location of the data */
			/* get it as a byte array type */
			tempArray.pointer = context->iInput + context->iIndex;
			tempArray.size = rawLength;

			/* Set the element value */
			cpiSetElementByteArrayValue(&rc, 
										element,
										&tempArray);
		}

		/* add this element as the last child */
		cpiAddAsLastChild(&rc, context->eCurrentElement, element);

		/* move on to the next item */
		context->iIndex += rawLength;

		/* count the tag processed */
		inTagCount++;
	}

	/* skip the delimiter */
	advance(context);

	/* finally, translate the rest of the message again, */
	/* since it may not have been translated initially.  */
	/* this is necessary since binary zeros in the raw   */
	/* data may stop the initial translation prematurely */
	cciMbsToUcs(&rc,
				(char *)context->iInput + context->iIndex,
				context->iBuffer + context->iIndex,
				context->iSize - context->iIndex,
				context->iCodePage);
#ifdef _DEBUG
	if (traceModule) 
	{
		Log("PLUGIN: <- processRawData(): name=%S\n", namePtr);
	}
#endif

}

/******************************************************************************/
/*                                                                            */
/* Routine to look up a variable by the FIX type and get the corresponding    */
/* metadata entry.                                                            */
/*                                                                            */
/******************************************************************************/

int findVarNameByType(FILETBL* fileTable, CciChar *type)

{
	int	i=0;

	while ((i < fileTable->maxvar) &&
		(CciCharCmp(type, fileTable->UnamePtr + fileTable->fvar[i].Utypeptr) != 0))
	{
		i++;
	}

	if (i == fileTable->maxvar)
	{
		i = -1;
	}

	return i;
}

/******************************************************************************/
/*                                                                            */
/* Routine to look up a variable by the assigned name and return the          */
/* corresponding metadata entry.                                              */
/*                                                                            */
/******************************************************************************/

int findVarByName(FILETBL* fileTable, CciChar *name)

{
	int	i=0;

	while ((i < fileTable->maxvar) &&
		(CciCharCmp(name, fileTable->UnamePtr + fileTable->fvar[i].Unameptr) != 0))
	{
		i++;
	}

	if (i == fileTable->maxvar)
	{
		i = -1;
	}

	return i;
}

/******************************************************************************/
/*                                                                            */
/* Routine to handle multi-string data types.                                 */
/*                                                                            */
/******************************************************************************/

void insertMultiStr(CciParser* parser, 
					PARSER_CONTEXT_ST* context, 
					CciChar *namePtr, 
					CciChar *startData, 
					CciChar *endData)

{
	int			markedSize;
	CciChar		*endSeg;

#ifdef _DEBUG
	if (traceModule) 
	{
		Log("PLUGIN: -> insertMultiStr(): name=%S, data=%S\n", namePtr, startData);
	}
#endif

	/* first, create a name value as a parent */
	createHeaderElement(parser, context, namePtr);

	/* now, break the data down into as many parts as needed */
	/* each item will be created with a name of Val */
	while (startData != endData)
	{
		/* find the end of the first data segment */
		/* data segments are separated by spaces */
		endSeg = startData;
		while ((endSeg != endData) && (CciCharIsSpace(endSeg[0]) != 0))
		{
			endSeg++;
		}

		/* calculate the data length */
		markedSize = endSeg - startData;

		/* add an element for this data item */
		/* use the defined data name */
		createVariableLengthElement(parser, context, MULTI_SEG_NAME, startData, markedSize);

		/* move on to the next item if needed */
		startData = endSeg;
		if (startData != endData)
		{
			/* find the next data item */
			while ((startData != endData) && (CciCharIsSpace(startData[0]) == 0))
			{
				startData++;
			}
		}
	}

	/* finally, finish the parent element */
	finishHeaderElement(context);

#ifdef _DEBUG
	if (traceModule) 
	{
		Log("PLUGIN: <- insertMultiStr(): name=%S, data=%S\n", namePtr, startData);
	}
#endif
}

/******************************************************************************/
/*                                                                            */
/* Routine to parse the elements using a metadata file that has been loaded   */
/* into memory.  This routine will parse one entry at a time.  The context    */
/* will be updated to reflect the new position within the message buffer.     */
/*                                                                            */
/******************************************************************************/

void parseElement(CciParser* parser, 
				  PARSER_CONTEXT_ST* context,
				  FILETBL* fileTable)

{
	int			rc;
	int			markedSize;
	int			typeLen;
	int			varnum;
	int			datType;
	unsigned char	flags;
	unsigned char	flags2;
	int			rptItem;
	CciChar		*startType;
	CciChar		*startData;
	CciChar		*namePtr;
	RPT_ITEM	*rptPtr;
	CciChar		name[MAX_NAME_LEN];
	char		charofs[32];
	char		errmsg[512];
#ifdef _DEBUG
	CciSize		size;
	CciChar		elemname[512];
	CciBool		completePrev;
	CciBool		completeNext;
#endif

#ifdef _DEBUG
	if (traceModule) 
	{
		completePrev = cpiElementCompletePrevious(&rc, context->eCurrentElement);
		completeNext = cpiElementCompleteNext(&rc, context->eCurrentElement);
		size = cpiElementName(&rc, context->eCurrentElement, elemname, sizeof(elemname) - 1); 
		Log("PLUGIN: -> parseElement(); element=0x%x %S bits(%d %d)\n",
            context->eCurrentElement, elemname, completePrev, completeNext);
	}
#endif

	/* first, isolate the next item */
	/* find the end of the variable type */
	startType = context->iBuffer + context->iIndex;
	typeLen = isolateTag(context);

	/* check if name length is zero */
	if (0 == typeLen)
	{
		/* throw an exception - zero length name */
		/* length field evaluates to zero */
		/* throw an exception             */
		/* Increase the invalid message counter */
		badMessageCount++;
		context->iBadMsg = 1;

		/* get the error message as a string */
		sprintf(errmsg, "No identifier found preceding equal sign at offset (%d)\n", 
			context->iIndex);

#ifdef _DEBUG
		if (traceModule) 
		{
			Log("%s\n", errmsg);
		}
#endif
		/* Write the error to the event viewer and terminate */
		sprintf(charofs, "%d", context->iIndex);
		cciThrowException(&rc, 
						  CCI_PARSER_EXCEPTION, 
						  __FILE__, 
						  __LINE__, 
						  "checkMessageFormat",
						  ERRDICT,
						  MISSING_IDENTIFIER,
						  errmsg,
						  charofs,
						  0);
	}

	/* check if we are in a repeating sequence at the moment */
	/* and if so, if this tag is also a member */
	rptPtr = context->iRepeat;
	while ((context->iRepeat != NULL) && 
		   (checkIfMember(context, startType) != 0) &&
		   (rptPtr != NULL))
	{
#ifdef _DEBUG
		if (1 == traceParser)
		{
			Log("context->iRepeat=%d, startType=%S\n", context->iRepeat, startType);
		}
#endif
		/* this item is no longer a member of the repeating sequence */
		/* finish the previous top level element */
		finishHeaderElement(context);

		/* remove it from the chain */
		rptPtr = context->iRepeat;
		context->iRepeat = (RPT_ITEM *) rptPtr->nextItem;

		/* free the storage */
		free(rptPtr);
	}

	/* look up the metadata and get the type of data and long name */
	varnum = findVarNameByType(fileTable, startType);

#ifdef _DEBUG
		if (1 == traceParser)
		{
			Log("varnum=%d, startType=%S\n", varnum, startType);
		}
#endif

	if (-1 == varnum)
	{
		/* didn't find the variable, so build a name from the type */
		datType = DATA_STRING;
		CciCharCpy(name, FIX_NAME_START);
		CciCharCpy(name + CciCharLen(name), startType);
		flags = 0;
		flags2 = 0;
		namePtr = name;
	}
	else
	{
		/* found the variable - get the name and type of data */
		datType = fileTable->fvar[varnum].Datatype;
		namePtr = fileTable->UnamePtr + fileTable->fvar[varnum].Unameptr;
		flags = fileTable->fvar[varnum].Flags;
		flags2 = fileTable->fvar[varnum].Flags2;
	}

#ifdef _DEBUG
		if (1 == traceParser)
		{
			Log("varnum=%d, startType=%S, name=%S\n", varnum, startType, namePtr);
		}
#endif

	/* check for a repeating item flag */
	if ((flags & FLAGS_REPEAT_FIRST) > 0)
	{
#ifdef _DEBUG
		if (1 == traceParser)
		{
			Log(" repeating item found name=%S\n", namePtr);
		}
#endif

		rptPtr = context->iRepeat;

		/* check if this item is another first in the same repeating sequence */
		if ((rptPtr != NULL) && 
			(CciCharCmp(rptPtr->typePtr, startType) == 0))
		{
			/* we have found another one of the same type */
#ifdef _DEBUG
			if (1 == traceParser)
			{
				Log(" found another new item in same structure name=%S\n", namePtr);
			}
#endif

			/* finish the current header */
			finishHeaderElement(context);

			/* create another header element */
			createHeaderElement(parser, context, rptPtr->topName);
		}
		else
		{
			/* this may be a repeating item about to start */
			/* check if out more carefully */
			rptItem = checkIfRpt(context, fileTable, startType, varnum);

#ifdef _DEBUG
			if (1 == traceParser)
			{
				Log(" rptItem=%d\n", rptItem);
			}
#endif

			/* check if this is the correct element to start a repeating sequence */
			if (rptItem >= 0)
			{
				/* we have a repeating item */
				/* allocate storage to keep track of this one */
				rptPtr = (RPT_ITEM *)malloc(sizeof(RPT_ITEM));

				/* initialize the new structure */
				rptPtr->nextItem = context->iRepeat;
				rptPtr->rptnum = rptItem;
				rptPtr->topName = fileTable->UnamePtr + fileTable->rvar[rptItem].Nameptr;
				rptPtr->typePtr = startType;
				rptPtr->members = fileTable->UnamePtr + fileTable->rvar[rptItem].ExtraTypes;

				/* insert it in the chain */
				context->iRepeat = rptPtr;

				/* create a header first    */
				createHeaderElement(parser, context, rptPtr->topName);
			}
		}
	}

#ifdef _DEBUG
	if (1 == traceParser)
	{
		Log(" datType=%d\n", datType);
	}
#endif

	if (DATA_RAW == datType)
	{
		/* This is raw data - we need to handle it differently */
		/* Raw data may contain any characters, including      */
		/* delimiters (0x01), string delimiters (0x00) and any */
		/* other character as well.                            */
		/* check if we have a length of a raw data type and */
		/* we need to decrypt this item                     */
		processRawData(parser, context, namePtr, flags, fileTable);
	}
	else
	{
		/* not raw data - we can handle it normally */
		/* isolate the data */
		/* point to the start of the data */
		startData = context->iBuffer + context->iIndex;

		/* find the SOH delimiter */
		markedSize = isolateData(context);

		/* terminate the data string by changing the delimiter */
		/* to a binary zero */
		startData[markedSize] = 0;

		/* check if we should ignore this item on input */
		if ((flags2 & FLAGS_SUPPRESS_IN) == 0)
		{
			/* check if we have a multiple string type */
			if (DATA_MULTI_STR == datType)
			{
				/* multiple string data type */
				insertMultiStr(parser, context, namePtr, startData, startData + markedSize);
			}
			else
			{
				/* build variable length entry */
				createVariableLengthElement(parser, context, namePtr, startData, markedSize);

				/* check if this is the MsgType field */
				if ((flags & FLAGS_MSGTYPE) > 0)
				{
					/* this is the message type field - capture the data */
					cciUcsToMbs(&rc, startData, context->sMsgType, MAX_FIX - 1, DEF_CODEPAGE);
				}
			}

			/* count this item */
			inTagCount++;
		}
	}

	/* remember the current element identifier, data and datatype */
	context->iPrevTag = startType;
	context->iPrevData  = startData;
	context->iPrevDataType = datType;

#ifdef _DEBUG
	if (traceModule) 
	{
		completePrev = cpiElementCompletePrevious(&rc, context->eCurrentElement);
		completeNext = cpiElementCompleteNext(&rc, context->eCurrentElement);
		size = cpiElementName(&rc, context->eCurrentElement, name, sizeof(name) - 1); 
		Log("PLUGIN: <- parseElement(); element=0x%x %S bits(%d %d)\n",
            context->eCurrentElement, name, completePrev, completeNext);
	}
#endif
}

/******************************************************************************/
/*                                                                            */
/* Plugin Parser Utility Function: Parse the next item                        */
/*                                                                            */
/******************************************************************************/

void parseNextItem(CciParser*  parser,
				   PARSER_CONTEXT_ST* context)

{
	int			rc;
	int			fileIdx=0;
	RPT_ITEM	*rptPtr;
#ifdef _DEBUG
	CciElement * body;
	CciSize size;
	CciChar name[512];
	CciBool	completePrev;
	CciBool completeNext;
#endif

#ifdef _DEBUG
	if (traceModule) 
	{
		completePrev = cpiElementCompletePrevious(&rc, context->eCurrentElement);
		completeNext = cpiElementCompleteNext(&rc, context->eCurrentElement);
		size = cpiElementName(&rc, context->eCurrentElement, name, sizeof(name) - 1); 
		Log("PLUGIN: -> parseNextItem(); currentElement=%S (0x%x) bits(%d %d) index=%d\n",
            name, context->eCurrentElement, completePrev, completeNext, context->iIndex);
	}
#endif

	/* find the metadata for this item */
#ifdef _DEBUG
	fileIdx = getMetaData(traceParser, context->sLevelName);
#else
	fileIdx = getMetaData(0, context->sLevelName);
#endif

	/* if the file was not found, then use the default entry */
	if (fileIdx < 0)
	{
		fileIdx = 0;
	}

	// Check if this is the first time through
	if (0 == context->iIndex)
	{
		// mark the body root first child flag as complete
		cpiSetElementCompletePrevious(&rc, context->eBody, COMPLETE_BIT);

#ifdef _DEBUG
		if (traceParser)
		{
			Log("Body element first child flag set to complete rc=%d body=0x%x\n", 
				rc, context->eBody);
		}
#endif
	}

	/* parse the next element in the message buffer */
	parseElement(parser, context, fileTable + fileIdx);

	/* Are we at the end of the buffer? */
	if (context->iIndex == context->iSize) 
	{
#ifdef _DEBUG
		if (traceParser) 
		{
			Log("PLUGIN: End of buffer reached...\n");
		}
#endif

		// check if we have any repeating items that have not been completed
		while (context->iRepeat != NULL)
		{
			// finish the header element
			finishHeaderElement(context);

			// unchain and free the storage
			rptPtr = context->iRepeat;
			context->iRepeat = (RPT_ITEM *)rptPtr->nextItem;
			free(rptPtr);
		}

		// set the body first child complete Next flag
		cpiSetElementCompleteNext(&rc, context->eCurrentElement, COMPLETE_BIT);

		// clear the current element, since we are done
		context->eCurrentElement = CCI_NULL_ADDR;
	}

#ifdef _DEBUG
	if (traceModule) 
	{
		// dump out the parse tree at this point
		body = cpiRootElement(&rc, parser);
		dumpAllElements(body, 0);
	}
#endif

#ifdef _DEBUG
	if (traceModule) 
	{
		completePrev = cpiElementCompletePrevious(&rc, context->eCurrentElement);
		completeNext = cpiElementCompleteNext(&rc, context->eCurrentElement);
		size = cpiElementName(&rc, context->eCurrentElement, name, sizeof(name) - 1); 
		Log("PLUGIN: <- parseNextItem() element=%S (0x%x) bits(%d %d), index %d\n", 
            name, context->eCurrentElement, completePrev, completeNext, context->iIndex);
	}
#endif


	return;
}

int calcOutputSize(CciElement *firstElement)

{
	int				rc;
	int				totLen=0;
	int				size;
	CciElementType	type;
	CciElement		*element;
	CciElement		*child;
	CciInt			tempInt;
	CciChar			UtempValue[32];
	CciByteArray	UtempByteArray;
	CciBool			tempBool;
	char			tempArray[32];
#ifdef _DEBUG
	CciChar			elemname[256];
#endif

#ifdef _DEBUG
	if (traceModule) 
	{
		size = cpiElementName(&rc, firstElement, elemname, sizeof(elemname) - 1); 
		Log("PLUGIN: -> calcOutputSize(); element=0x%x %S\n",
            firstElement, elemname);
	}
#endif

	/* get the body element*/
	element = firstElement;

	/* get the first child of the body */
	while (element != NULL)
	{
		/* allow for a maximum sized type of 5 digits */
		/* plus an equal sign and a delimiter         */
		totLen += 7;

		/* calculate the length of the data */
		UtempValue[sizeof(UtempValue) - 1] = 0;
		UtempValue[0] = 0;

		/* first, get the element type      */
		type = cniElementValueType(&rc, element);
		switch (type)
		{
		case CCI_VALUE_TYPE_BOOL:
			{
				/* will take one character for output */
				tempBool = cpiElementBooleanValue(&rc, element);
				switch(tempBool)
				{
				case TRUE:
					{
						UtempValue[0] = 'Y';
						UtempValue[1] = 0;
					}
				case FALSE:
					{
						UtempValue[0] = 'N';
						UtempValue[1] = 0;
					}
				case UNKNOWN:
					{
						UtempValue[0] = '?';
						UtempValue[1] = 0;
					}
				default:
					{
						UtempValue[0] = 0;
					}
				}

				totLen ++;
				break;
			}
		case CCI_VALUE_TYPE_INTEGER:
			{
				tempInt = cpiElementIntegerValue(&rc, element);

#ifdef WIN32
				sprintf(tempArray, "%I64d", tempInt);
#else
				sprintf(tempArray, "%lld", tempInt);
#endif

				totLen += strlen(tempArray);;
				break;
			}
		case CCI_VALUE_TYPE_CHAR:
			{
				size = 0;
				size = cpiElementCharacterValue(&rc, element, UtempValue, sizeof(UtempValue) - 1);
				totLen += size;
				break;
			}
		case CCI_VALUE_TYPE_BYTEARRAY:
			{
				/* find out how big the array is */
				UtempByteArray.size = 1;
				UtempByteArray.pointer = tempArray;
				size = cpiElementByteArrayValue(&rc, element, &UtempByteArray);
				totLen += size;				
				break;
			}
		}

		/* handle any children of this element */
		child = cpiFirstChild(&rc, element);
		if ((CCI_SUCCESS == rc) && (child != CCI_NULL_ADDR))
		{
			totLen += calcOutputSize(child);
		}

#ifdef _DEBUG
		if (traceWrite) 
		{
			size = cpiElementName(&rc, element, elemname, sizeof(elemname) - 1); 
			switch (type)
			{
			case CCI_VALUE_TYPE_CHAR:
			case CCI_VALUE_TYPE_BOOL:
				{
					Log(" element=0x%x %S valueType %d value %S totLen=%d\n",
				        element, elemname, type, UtempValue, totLen);
					break;
				}
			case CCI_VALUE_TYPE_INTEGER:
				{
					Log(" element=0x%x %s valueType %d value %S totLen=%d\n",
				        element, elemname, type, tempArray, totLen);
					break;
				}
			default:
				{
					Log(" element=0x%x %S valueType %d totLen=%d\n",
				        element, elemname, type, totLen);
					break;
				}
			}
		}
#endif

		/* move on to the next element */
		element = cpiNextSibling(&rc, element);
	}

#ifdef _DEBUG
	if (traceModule) 
	{
		Log("PLUGIN: <- calcOutputSize() returns totLen=%d\n", totLen);
	}
#endif

	return totLen;
}

int createMultiString(PARSER_CONTEXT_ST* context, CciElement *element)

{
	int				rc;
	int				len=0;
	int				tempLen;
	CciElement		*child;
#ifdef _DEBUG
	int				size;
	CciChar			elemname[256];
#endif

#ifdef _DEBUG
	if (traceModule) 
	{
		size = cpiElementName(&rc, element, elemname, sizeof(elemname) - 1); 
		Log("PLUGIN: -> createMultiString(); element=0x%x %S\n", element, elemname);
	}
#endif

	/* find the first child */
	child = cpiFirstChild(&rc, element);

	/* process all the children, gathering their values */
	while ((CCI_SUCCESS == rc) && (CCI_NULL_ADDR != child))
	{
		/* insert this value */
		rc = findCharacterValue((char *)context->iOutput + context->iIndex,
								context->iRemaining,
								child,
								context->iCodePage);

		/* move the output pointer and remember remaining buffer */
		tempLen = strlen((char *)context->iOutput + context->iIndex);
		context->iIndex += tempLen;

		/* now insert a blank as a delimiter */
		cciUcsToMbs(&rc,
					STR_SPACE,
					(char *)context->iOutput + context->iIndex,
					1,
					context->iCodePage);

		/* keep track of the length and the remaining bytes */
		len += tempLen + 1;
		context->iRemaining -= tempLen + 1;
		context->iIndex++;

		/* move on to the next child variable */
		child = cpiNextSibling(&rc, child);
	}

	// get rid of the last space
	if (len > 0)
	{
		context->iOutput[context->iIndex--] = 0;
		context->iRemaining++;
		len--;
	}

#ifdef _DEBUG
	if (traceModule) 
	{
		Log("PLUGIN: <- createMultiString() returns length %d\n", len);
	}
#endif

	return len;
}

/******************************************************************************/
/*                                                                            */
/* Routine to look up a variable in the variable table based on element name. */
/*                                                                            */
/******************************************************************************/

int lookupVariable(CciChar *name, FILETBL *fileTable)

{
	int		varnum=-1;

#ifdef _DEBUG
	if (traceModule) 
	{
		Log("PLUGIN: -> lookupVariable(); name=%S\n", name);
	}
#endif

	/* what type of name is this? */
	if (memcmp(name, FIX_NAME_START, sizeof(FIX_NAME_START) - sizeof(CciChar)) == 0)
	{
		/* the name contains the type */
		/* look up the variable by type */
		varnum = findVarNameByType(fileTable, name + 2);
	}
	else
	{
		/* look up the name to find the type */
		varnum = findVarByName(fileTable, name);
	}

#ifdef _DEBUG
	if (traceModule) 
	{
		Log("PLUGIN: <- lookupVariable(); varnum=%d\n", varnum);
	}
#endif

	return varnum;
}

/******************************************************************************/
/*                                                                            */
/* Routine to convert a variable name into an identifier.                     */
/*                                                                            */
/******************************************************************************/

void getIdentifier(FILETBL *fileTable, 
				   CciChar *name, 
				   char * tempType)

{
	int			rc;
	int			size;
	int			varnum;

#ifdef _DEBUG
	if (traceModule) 
	{
		Log("PLUGIN: -> getIdentifier(); name=%S\n", name);
	}
#endif

	/* initialize the output area */
	tempType[0] = 0;

	/* see if we can find this variable */
	varnum = lookupVariable(name, fileTable);

	if (varnum >= 0)
	{
		/* get the variable type in a work area */
		size = cciUcsToMbs(&rc,
						   fileTable->UnamePtr +  + fileTable->fvar[varnum].Utypeptr,
						   tempType,
						   MAX_TAG_LEN,
						   DEF_CODEPAGE);
	}
	else
	{
		if (memcmp(name, FIX_NAME_START, sizeof(FIX_NAME_START) - sizeof(CciChar)) == 0)
		{
			/* get the type of this variable */
			/* move the type to the output area */
			size = cciUcsToMbs(&rc,
							   name + 2,
							   tempType,
							   MAX_TAG_LEN,
							   DEF_CODEPAGE);
		}
	}

#ifdef _DEBUG
	if (traceModule) 
	{
		Log("PLUGIN: <- getIdentifier(); identifier=%s\n", tempType);
	}
#endif
}

/******************************************************************************/
/*                                                                            */
/* Routine to create the first two elements in the header.                    */
/*                                                                            */
/* The first two elements are the BeginString (type #8) and MessageLength     */
/*  (type #9).  The BeginString contains a version string.  If BeginString    */
/*  element is found at the beginning of the logical message tree, then the   */
/*  element value will be used as the version.  Otherwise, a default value    */
/*  for version 4.2 will be used.  The identifier and equal sign for the      */
/*  MessageLength field will be filled in but the actual length will be       */
/*  added after the rest of the message is built.                             */
/*                                                                            */
/******************************************************************************/

void createHeader(PARSER_CONTEXT_ST* context, char *tempHeader)

{
	int				i;
	int				j;
	int				rc;
	int				size;
	int				len;
	char			value[32];
	CciChar			name[256];

	/* get the name of the first variable */
	size = cpiElementName(&rc, context->eCurrentElement, name, sizeof(name) - 1);

#ifdef _DEBUG
	if (traceModule) 
	{
		Log("PLUGIN: -> createHeader(); element=0x%x %S\n", context->eCurrentElement, name);
	}
#endif

	/* initialize the output area to a BeginString identifier */
	size = cciUcsToMbs(&rc, BEGIN_STRING, tempHeader, sizeof(tempHeader) - 1, context->iCodePage);

	/* check if we found a BeginString variable (type #8) */
	if ((CciCharCmp(name, BEGINSTRING) == 0) || (CciCharCmp(name, FX8) == 0))
	{
		/*found it - get the value from the parse tree */
		rc = findCharacterValue(tempHeader + strlen(tempHeader),
								16,
								context->eCurrentElement,
								context->iCodePage);

		/* initialize the value area to nulls */
		memset(value, 0, sizeof(value));

		/* get the value of this element */
		rc = findCharacterValue(value, 
								sizeof(value) - 1, 
								context->eCurrentElement,
								DEF_CODEPAGE);

		/* get the value as the level, with any periods removed */
		i = 0;
		j = 0;
		while ((value[i] != 0) && (j < (sizeof(context->sLevelName) - 1)))
		{
			/* check for a period */
			if (value[i] != chPeriod)
			{
				context->sLevelName[j++] = value[i++];
			}
			else
			{
				i++;
			}
		}

		/* terminate the string */
		context->sLevelName[j] = 0;

		/* move on to the next element */
		context->eCurrentElement = cpiNextSibling(&rc, context->eCurrentElement);

		/* get the name of the next element */
		size = cpiElementName(&rc, context->eCurrentElement, name, sizeof(name) - 1);
	}
	else
	{
		/* No BeginString or FX8 variable  - use a default value */
		size = cciUcsToMbs(&rc,
						   DEFAULT_VERSION,
						   tempHeader + strlen(tempHeader),
						   16,
						   context->iCodePage);

		/* fill in the default level of metadata */
		strcpy(context->sLevelName, DEFAULT_LEVEL);
	}

	/* insert the delimiter and message length identifier */
	len = strlen(tempHeader);
	tempHeader[len++] = DELIMITER;
	size = cciUcsToMbs(&rc,
					   MESSAGE_LENGTH,
					   tempHeader + len,
					   16,
					   context->iCodePage);

	/* check if we have a length field next */
	if ((CciCharCmp(name, BODYLENGTH) == 0) ||  (CciCharCmp(name, FX9) == 0))
	{
		/* skip the length element */
		context->eCurrentElement = cpiNextSibling(&rc, context->eCurrentElement);
	}

#ifdef _DEBUG
	if (traceModule) 
	{
		size = cpiElementName(&rc, context->eCurrentElement, name, sizeof(name) - 1); 
		Log("PLUGIN: <- createHeader(); element=0x%x %S tempHeader %s\n", 
			context->eCurrentElement, name, tempHeader);
	}
#endif
}

int createOutput(PARSER_CONTEXT_ST* context, 
				 CciElement *firstElement, 
				 FILETBL *fileTable)

{
	int				rc=CCI_SUCCESS;
	int				size;
	int				varnum;
	int				datType;
	int				maxvar;
	int				len;
	int				saveIndex;
	int				saveLen;
	int				encryptLen;
	unsigned char	*newOutput;
	VAR_FILE_REC	*fvar;
	CciElement		*element;
	CciChar			*tagPtr=CCI_NULL_ADDR;
	CciChar			*tagLengthPtr=CCI_NULL_ADDR;
	CciElementType	elemtype;
	CciChar			name[256];
	CciChar			UtempLen[32];
	char			tempLen[32];
	char			tempName[256];
	unsigned char	flags=0;
	unsigned char	flags2=0;
	char			errmsg[512];

#ifdef _DEBUG
	if (traceModule) 
	{
		size = cpiElementName(&rc, firstElement, name, sizeof(name) - 1); 
		Log("PLUGIN: -> createOutput(); element=0x%x %S\n", firstElement, name);
	}
#endif

	/* set two local variables for better efficiency */
	maxvar = fileTable->maxvar;
	fvar = fileTable->fvar;

	/* process the rest of the parse tree */
	element = firstElement;

	/* loop through all the elements */
	while ((CCI_SUCCESS == rc) && (element != CCI_NULL_ADDR) && (context->iRemaining > 0))
	{
		/* we need to see if we can find meta data for this element */
		/* get the name of this element */
		size = cpiElementName(&rc, element, name, sizeof(name) - 1);

		/* find the metadata entry for this element */
		varnum = lookupVariable(name, fileTable);

		/* check if we found the metadata */
		if (varnum >= 0)
		{
			/* get the flags */
			flags = fileTable->fvar[varnum].Flags;
			flags2 = fileTable->fvar[varnum].Flags2;
			tagPtr = fileTable->UnamePtr + fileTable->fvar[varnum].Utypeptr;
			tagLengthPtr = fileTable->UnamePtr + fileTable->fvar[varnum].Ulentype;
		}

		/* get the element type */
		elemtype = cpiElementType(&rc, element);

		/* is this a name only element? */
		if (CCI_ELEMENT_TYPE_NAME == elemtype)
		{
			/* this may be a group of data that is to be encrypted */
			/* check if metadata indicates encryption " */
#ifdef _DEBUG
			if (traceModule) 
			{
				Log("Found name only element %S remaining %d\n", name, context->iRemaining);
			}
#endif

			/* remember where we are starting within the output */
			saveIndex = context->iIndex;

			/* create the output */
			createOutput(context, cpiFirstChild(&rc, element), fileTable);

			/* remember how long the clear data is */
			saveLen = context->iIndex - saveIndex;

			/* check if this data needs to be encrypted */
			if (((flags & FLAGS_DECRYPT_PARSE) > 0) && 
				(context->iUseEncryption != 0))
			{
				/* this data is supposed to be encrypted */
				/* allocate a new area for the encrypted data */
				newOutput = (unsigned char *)malloc(context->iLength);

				/* get the encrypted data and the length of the encrypted data */
				encryptLen = encrypt(context->iOutput + saveIndex, 
									 saveLen, 
									 newOutput);

				/* go back to the right location in the output buffer */
				context->iIndex = saveIndex;

				/* create a secure length item tag */
				cciUcsToMbs(&rc, 
							tagLengthPtr, 
							(char *)context->iOutput + context->iIndex,
							context->iRemaining,
							context->iCodePage);
				context->iIndex += CciCharLen(tagLengthPtr);
				context->iRemaining -= CciCharLen(tagLengthPtr);

				/* insert an equal sign as a delimiter */
				insertEqual(context);

				/* get the length as a character value in the desired code page */
				sprintf(tempLen, "%d", encryptLen);
				cciMbsToUcs(&rc, tempLen, UtempLen, sizeof(tempLen) - 1, DEF_CODEPAGE);
				cciUcsToMbs(&rc,
							UtempLen,
							(char *)context->iOutput + context->iIndex,
							context->iRemaining,
							context->iCodePage);
				context->iIndex += strlen(tempLen);

				/* now insert a data delimiter character */
				context->iOutput[context->iIndex++] = DELIMITER;
				context->iRemaining -= strlen(tempLen) + 1;

				/* count the item */
				outTagCount++;

				/* create the secure data item tag */
				cciUcsToMbs(&rc, 
							tagPtr, 
							(char *)context->iOutput + context->iIndex,
							context->iRemaining,
							context->iCodePage);
				context->iIndex += CciCharLen(tagPtr);;
				context->iRemaining -= CciCharLen(tagPtr);

				/* insert an equal sign as a delimiter */
				insertEqual(context);

				/* now insert the encrypted data */
				memcpy(context->iOutput + context->iIndex, newOutput, encryptLen);
				context->iIndex += encryptLen;

				/* account for any difference in length */
				context->iRemaining -= encryptLen - saveLen;

				/* finally insert a data delimiter character */
				context->iOutput[context->iIndex++] = DELIMITER;
				context->iRemaining--;

				/* release the storage */
				free(newOutput);

				/* count the item */
				outTagCount++;
			}
		}
		else
		{
			if (varnum >= 0)
			{
				/* found the data element */
				/* insert the name into the output stream */
				len = strlen(fileTable->namePtr + fvar[varnum].Typeptr);
				if (len < context->iRemaining)
				{
					/* move the type to the output area */
					size = cciUcsToMbs(&rc,
									   fileTable->UnamePtr + fvar[varnum].Utypeptr,
									   (char *)context->iOutput + context->iIndex,
									   context->iRemaining,
									   context->iCodePage);
			
					context->iRemaining -= len;
					context->iIndex += len;
				}

				/* get the data type */
				datType = fvar[varnum].Datatype;
			}
			else
			{
				/* force the data type to string */
				datType = DATA_STRING;

				/* unable to locate any metadata for this element */
				/* check if the data type is contained in the name */
				if (memcmp(name, FIX_NAME_START, sizeof(FIX_NAME_START) - sizeof(CciChar)) == 0)
				{
					/* the type appears to be in the element name */
					context->iRemaining -= size;
					context->iIndex += size;
				}
				else
				{
					/* Do not recognize the data name - throw an exception */
					/* count the rejected output message */
					badOutputCount++;
					context->iBadMsg = 1;

					/* translate the name to local code page */
					cciUcsToMbs(&rc, name, tempName, sizeof(tempName) - 1, DEF_CODEPAGE);

					sprintf(errmsg, "Unrecognized data name in parse tree - %s\n", tempName);

#ifdef _DEBUG
					if (traceWrite > 0)
					{
						Log(errmsg);
					}
#endif
					/* throw a parser exception */
					cciThrowException(&rc, 
									  CCI_PARSER_EXCEPTION, 
									  __FILE__, 
									  __LINE__, 
									  "createOutput",
									  ERRDICT,
									  INVALID_ELEMENT_NAME,
									  errmsg,
									  tempName,
									  0);
				}
			}

			/* now insert an equal sign */
			if (context->iRemaining > 0)
			{
				insertEqual(context);
			}

			/* append the data and delimiter for this item */
			if (datType != DATA_MULTI_STR)
			{
				/* normal name and value item */
				rc = findCharacterValue((char *)context->iOutput + context->iIndex,
										context->iRemaining,
										element,
										context->iCodePage);

				/* figure out how many characters of output we created */
				context->iRemaining -= strlen((char *)context->iOutput + context->iIndex);
				context->iIndex += strlen((char *)context->iOutput + context->iIndex);
			}
			else
			{
				/* multiple string item */
				len = createMultiString(context, element);
			}

			/* add a delimiter */
			if (context->iRemaining > 0)
			{
				context->iOutput[context->iIndex++] = DELIMITER;
				context->iRemaining--;
			}

			/* count the item */
			outTagCount++;
		}

#ifdef _DEBUG
		if (traceModule) 
		{
			Log("Data appended for element %S next offset %d remaining %d\n", 
				name, context->iIndex, context->iRemaining);
		}
#endif

		/* check if there are any children of this element */
		/* go on to the next element */
		element = cpiNextSibling(&rc, element);
	}	/* while */

#ifdef _DEBUG
	if (traceModule) 
	{
		Log("PLUGIN: <- createOutput() created buffer of %d bytes\n", context->iIndex);
	}
#endif

	return context->iIndex;
}

int createOutputMessage(CciParser* parser, PARSER_CONTEXT_ST* context)

{
	int			rc;
	int			i;
	int			length=0;
	int			storageLen=0;
	int			totalLen=0;
	int			fileIdx;
	int			hdrLen;
	int			checksum1;
	int			checksum2;
	char		tempHeader[64];
	char		tempchLen[32];
	CciChar		UtempchLen[32];
	char		errmsg[512];
#ifdef _DEBUG
	int			size;
	CciChar		elemname[256];
#endif

#ifdef _DEBUG
	if (traceModule) 
	{
		size = cpiElementName(&rc, context->eCurrentElement, elemname, sizeof(elemname) - 1); 
		Log("PLUGIN: -> createOutputMessage(); element=0x%x %S\n", context->eCurrentElement, elemname);
		dumpAllElements(context->eBody, 0);
	}
#endif

	/* get the first child of the body */
	context->eCurrentElement = cpiFirstChild(&rc, context->eBody);

	/* create the first part of the header */
	createHeader(context, (char *)tempHeader);
	hdrLen = strlen(tempHeader);

	/* see if we can find a metadata file to match */
#ifdef _DEBUG
	fileIdx = getMetaData(traceWrite, context->sLevelName);
#else
	fileIdx = getMetaData(0, context->sLevelName);
#endif

	if (fileIdx < 0)
	{
		/* throw an exception if can't find metadata for this output message */
		badOutputCount++;
		context->iBadMsg = 1;

		sprintf(errmsg, "No metadata found for this fix level - %s\n", context->sLevelName);

#ifdef _DEBUG
		if (traceWrite > 0)
		{
			Log(errmsg);
		}
#endif
		/* throw a parser exception */
		cciThrowException(&rc, 
						  CCI_PARSER_EXCEPTION, 
						  __FILE__, 
						  __LINE__, 
						  "createOutputMessage",
						  ERRDICT,
						  INVALID_FIX_LEVEL,
						  errmsg,
						  context->sLevelName,
						  0);
	}

	/* get the total length of the output data buffer required */
	/* allocate a bit more than we think we need */
	context->iLength = calcOutputSize(context->eCurrentElement) + 256;

	/* allocate storage for an output buffer */
	context->iOutput = (unsigned char *)malloc(context->iLength);
	context->iRemaining = context->iLength;

	/* create the body of the output message */
	length = createOutput(context, context->eCurrentElement, &fileTable[fileIdx]);

	/* finish the header */
	/* get the length as a string in local code page */
	/* and then translate to the desired output code page */
	sprintf(tempchLen, "%d", length);
	cciMbsToUcs(&rc, tempchLen, UtempchLen, sizeof(tempchLen) - 1, DEF_CODEPAGE);
	cciUcsToMbs(&rc,
				UtempchLen,
				tempHeader + hdrLen,
				hdrLen,
				context->iCodePage);

	/* update the length and add a delimiter */
	hdrLen += strlen(tempchLen);
	tempHeader[hdrLen++] = DELIMITER;
	tempHeader[hdrLen] = 0;

	/* append the header to the buffer */
	cpiAppendToBuffer(&rc, parser, (CciByte *) tempHeader, hdrLen);

	/* calculate the checksum for the total message */
	checksum1 = calcCheckSum((unsigned char *)tempHeader, hdrLen);
	checksum2 = calcCheckSum(context->iOutput, context->iIndex);
	checksum1 = (checksum1 + checksum2) % 256;

	/* get the checksum as a character string */
	memset(tempchLen, 0, sizeof(tempchLen));
	sprintf(tempchLen, "%-.3d", checksum1);

	/* convert any blanks to charcter zeros */
	i = 0;
	while (tempchLen[i] != 0)
	{
		if (' ' == tempchLen[i])
		{
			tempchLen[i] = '0';
		}

		i++;
	}

	/* add the checksum entry to the end of the body */
	cciUcsToMbs(&rc,
				CHECK_SUM,
				(char *)context->iOutput + context->iIndex,
				CHECK_SUM_LEN,
				context->iCodePage);
	context->iIndex += CHECK_SUM_LEN;

	/* insert the check sum following the checksum tag */
	cciMbsToUcs(&rc, tempchLen, UtempchLen, 3, DEF_CODEPAGE);
	cciUcsToMbs(&rc,
				UtempchLen,
				(char *)context->iOutput + context->iIndex,
				3,
				context->iCodePage);
	context->iIndex += 3;

	/* insert the delimiter at the end of the checksum entry */
	context->iOutput[context->iIndex++] = DELIMITER;

	/* append the result to the buffer */
	cpiAppendToBuffer(&rc, parser, (CciByte *) context->iOutput, context->iIndex);

	/* release the storage we acquired */
	free(context->iOutput);
	context->iOutput = NULL;

#ifdef _DEBUG
	if (traceModule) 
	{
		Log("PLUGIN: <- createOutputMessage() appended %d bytes - remaining was %d\n", 
			context->iIndex + hdrLen, context->iRemaining);
	}
#endif

	return context->iIndex + hdrLen;
}

const char * getFixSubLevel()

{
	return Level;
}