//  mtxstruc.h 

/*
 * Licensed Materials - Property of IBM
 * 5648-C63
 * (C) Copyright IBM Corp. 2000
 */

/*****************************************************************/
/*                                                               */
/*  Definitions for mtx file layout.                             */
/*                                                               */
/*****************************************************************/

#define MAX_NAME_LEN	63
#define META_HDR_LEN	32
#define REPEAT_ITEM_LEN	32
#define SEG_HDR_LEN		20

/*****************************************************************/
/*                                                               */
/*  Data type definitions        .                               */
/*                                                               */
/*****************************************************************/

#define DATA_INTEGER	0
#define DATA_FLOAT		1
#define DATA_QTY		2
#define DATA_PRICE		3
#define DATA_PR_OFFSET	4
#define DATA_AMT		5
#define DATA_CHAR		6
#define DATA_BOOL		7
#define DATA_STRING		8
#define DATA_MULTI_STR	9
#define DATA_CURRENCY	10
#define DATA_EXCHANGE	11
#define DATA_TIMESTAMP	12
#define DATA_TIMEONLY	13
#define DATA_DATEONLY	14
#define DATA_MARKET_DT	15
#define DATA_RAW		16
#define DATA_YYYYMM		17
#define DATA_DAY		18

/*****************************************************************/
/*                                                               */
/*  Flag bit definitions         .                               */
/*                                                               */
/*****************************************************************/

#define FLAGS_REPEAT_COUNT	1		/* Repeating item count field */
#define FLAGS_REPEAT_FIRST	2		/* First field in repeating sequence */
#define FLAGS_RAW_LENGTH	4		/* Raw data length field */
#define FLAGS_LENGTH		8		/* Message length parameter */
#define FLAGS_MSGTYPE		16		/* Flag to indicate MsgType field */
#define FLAGS_DECRYPT_PARSE	32		/* Decrypt and parse this field on input */
#define FLAGS_DECRYPT_ONLY	64		/* Decrypt this field on input */
#define FLAGS_SUPPRESS_IN	1		/* Do not include this in input tree */
#define FLAGS_SUPPRESS_OUT	2		/* Do not include this in output */

/**************************************************************/
/*                                                            */
/* Metadata file header structure                             */
/*                                                            */
/**************************************************************/

typedef struct {
	char	HdrLength;
	char	Version;
	char	Resvd[2];
	int		Maxvar;
	int		VarTable;
	int		Names;
	int		Unames;
	int		RepeatArea;
	int		RepeatCount;
	char	Resvd2[4];
} META_FILE_HEADER;

/**************************************************************/
/*                                                            */
/* Repeating item structure definition                        */
/*                                                            */
/**************************************************************/

typedef struct {
	char	Name[8];
	int		Nameptr;
	int		Unameptr;
	int		FirstUtypeptr;
	int		CountUtypeptr;
	int		ExtraTypes;
	char	Resvd[4];
} REPEAT_REC;

/**************************************************************/
/*                                                            */
/* Variable table structure definition                        */
/*                                                            */
/**************************************************************/

typedef struct {
	int		Nameptr;
	int		Unameptr;
	int		Typeptr;
	int		Utypeptr;
	int		Ulentype;
	char	Datatype;
	unsigned char	Flags;
	unsigned char	Flags2;
	char	Resvd;
} VAR_FILE_REC;
