/**************************************************************/
/*                                                            */
/* Program Name:  PRINTMTX.C                                  */
/*                                                            */
/*  Program to print the contents of an mtx data file.        */
/*                                                            */
/*  Written by Jim MacNair - April, 2000                      */
/*                                                            */
/**************************************************************/

/**************************************************************/
/*                                                            */
/*  Current Version 1.00                                      */
/*                                                            */
/**************************************************************/

/**************************************************************/
/* Required include for C functions                           */
/**************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

/**************************************************************/
/*                                                            */
/* Defines and global variables                               */
/*                                                            */
/**************************************************************/

#include "mtxstruc.h"


static char copyright[] = "(C) Copyright IBM Corp, 2001";
static char Version[]=\
"@(#)printmtx V1.00 - program to print fix metadata files  - Jim MacNair ("__DATE__")";

/*******************************************************************/
/*                                                                 */
/* Routine to convert a data type to the coded value.              */
/*                                                                 */
/*******************************************************************/

char * getDataType(int dtype)

{
	char	*ptr;

	switch(dtype)
	{
	case DATA_INTEGER:
		{
			ptr = "Integer";
			break;
		}
	case DATA_FLOAT:
		{
			ptr = "Float";
			break;
		}
	case DATA_QTY:
		{
			ptr = "Quantity";
			break;
		}
	case DATA_PRICE:
		{
			ptr = "Price";
			break;
		}
	case DATA_PR_OFFSET:
		{
			ptr = "Price Ofs";
			break;
		}
	case DATA_AMT:
		{
			ptr = "Amount";
			break;
		}
	case DATA_CHAR:
		{
			ptr = "Char";
			break;
		}
	case DATA_BOOL:
		{
			ptr = "Bool";
			break;
		}
	case DATA_STRING:
		{
			ptr = "String";
			break;
		}
	case DATA_MULTI_STR:
		{
			ptr = "Multi Str";
			break;
		}
	case DATA_CURRENCY:
		{
			ptr = "Currency";
			break;
		}
	case DATA_EXCHANGE:
		{
			ptr = "Exchange";
			break;
		}
	case DATA_TIMESTAMP:
		{
			ptr = "Timestamp";
			break;
		}
	case DATA_DATEONLY:
		{
			ptr = "Date Only";
			break;
		}
	case DATA_TIMEONLY:
		{
			ptr = "Time Only";
			break;
		}
	case DATA_MARKET_DT:
		{
			ptr = "Market Dt";
			break;
		}
	case DATA_RAW:
		{
			ptr = "Raw";
			break;
		}
	case DATA_YYYYMM:
		{
			ptr = "YYYYMM";
			break;
		}
	case DATA_DAY:
		{
			ptr = "Day of Mon";
			break;
		}
	default:
		{
			ptr = "";
			break;
		}
	}
	
	return ptr;
}
/**************************************************************/
/*                                                            */
/* Routine to write the input file name to the message file.  */
/*                                                            */
/**************************************************************/

void printReport(VAR_FILE_REC *fvar,
				char * nameArea,
				wchar_t * UnameArea,
				const int maxvar, 
				FILE *msgfile)

{
	int i;
	char* namePtr;
	char* typePtr;
	char* dataTypePtr;
	wchar_t *UrawLengthPtr;

	// Print a heading
	fprintf(msgfile," Type Format    RawLen   Flags   Name\n\n");

	// Print the individual lines
	for (i=0; i < maxvar; i++)
	{
		// point to the name in the name area
		namePtr = nameArea + fvar[i].Nameptr;
		typePtr = nameArea + fvar[i].Typeptr;
		dataTypePtr = getDataType(fvar[i].Datatype);
		UrawLengthPtr = UnameArea + fvar[i].Ulentype;

		fprintf(msgfile, "%5.5s %-10.10s %5.5S 0x%.2x 0x%.2x %s\n",
				typePtr,
				dataTypePtr,
				UrawLengthPtr,
				fvar[i].Flags,
				fvar[i].Flags2,
				namePtr);
	}

	return;
}

/**************************************************************/
/*                                                            */
/* Routine to write the repeating fields to the message file. */
/*                                                            */
/**************************************************************/

void printRepeats(REPEAT_REC *rvar, wchar_t *UnameArea, int maxrpt, FILE *msgfile)

{
	int		i=0;
	char	*ptr;
	wchar_t	*countTag;
	wchar_t	*firstTag;
	wchar_t	*nameptr;
	wchar_t *nextitem;
	wchar_t items[2048];

	/* print a header */
	fprintf(msgfile, "\n\n\nRepeating item definitions\n");
	fprintf(msgfile, "\nMsg Cnt 1st  name          Allowed items\n\n");

	while (i < maxrpt)
	{
		/* get the repeat count tag id */
		ptr = (char *) UnameArea;
		countTag = UnameArea + rvar[i].CountUtypeptr;
		firstTag = UnameArea + rvar[i].FirstUtypeptr;
		nameptr  = UnameArea + rvar[i].Nameptr;
		nextitem = UnameArea + rvar[i].ExtraTypes;

		/* initialize to a null string */
		items[0] = 0;

		while (nextitem[0] != 0)
		{
			wcscat(items, nextitem);
			wcscat(items, L" ");
			nextitem += wcslen(nextitem) + 1;
		}

		fprintf(msgfile, "%2S %3S  %3S  %-20S %S\n", rvar[i].Name, countTag, firstTag, nameptr, items);
		i++;
	}
}

/**************************************************************/
/*                                                            */
/* Routine to write the input file name to the message file.  */
/*                                                            */
/**************************************************************/

void writeFileName(char *fileName, FILE *msgfile)

{
	/* write the file name to the message file */
	fprintf(msgfile, "Processing file %s\n", fileName);
}

/**************************************************************/
/*                                                            */
/* Routine to display the command syntax.                     */
/*                                                            */
/**************************************************************/

void printHelp(char *cmd)
{
   printf("\nPRINTMTX V1.00\n");
   printf("\nCommand usage is:\n\n");
   printf("  %s mtx-file-name\n\n", cmd);
   printf("A report file is printed with with the name of\n");
   printf("the metadata file and an extension of msg\n");
}

/**************************************************************/
/*                                                            */
/*                                                            */
/*                                                            */
/**************************************************************/

void main(int argc, char *argv[])

{
	int	dataLen=0;			// length of COBOL data               
	VAR_FILE_REC * fvar;	// Pointer to metadata file variable definitions 
	REPEAT_REC * rvar;		// Pointer to metadata repeat field definitions
	META_FILE_HEADER* hdr;	// Pointer to metadata file header area
	char *ptr;				// work variable
	char *nameArea;			// Pointer to metadata file name area
	wchar_t *UnameArea;
	char *fileExt;			// file extension                     
	char * tempPtr;			// temporary pointer
	char msgFileName[255];	// name of message file               
	char mtxFileName[255];	// name of mtx file                   

 /*
  * File variables and counters
  */

    FILE        *mtxfile;	// input metadata file         
    FILE        *msgfile;	// conversion program messages 

/**************************************************************/
/*     Allocate and initialize variable areas                 */
/**************************************************************/

	memset(mtxFileName, 0, sizeof(mtxFileName));
	memset(msgFileName, 0, sizeof(msgFileName));

/**************************************************************/
/*     Parse the file name and generate new file name.        */
/**************************************************************/

	if (argc > 1)        // check for file name 
	{
		// Get the file name parameter, making sure not to overflow 
		strncpy(mtxFileName,argv[1],sizeof(mtxFileName) - 1);

		// check for help request 
		if ((mtxFileName[0] == '?') || (mtxFileName[1] == '?'))
		{
			printHelp(argv[0]);
			exit(1);
		} // endif 
	}
	else
	{
		// must have file name for mtx file 
		printHelp(argv[0]);
		exit(1);
	} // endif 


/**************************************************************/
/* Generate the new file name                                 */
/**************************************************************/

	// get the full file name 
	strcpy(msgFileName,mtxFileName);

	// Remove any extension from the message file name 
	// look for a period 
	fileExt = strchr(msgFileName, '.');
	if (fileExt != NULL)
	{
		// remove the file extension if it exists 
		(*fileExt) = 0;
	} // endif 

	// add a file extension to the message files 
	strcat(msgFileName,".msg");

/**************************************************************/
/* Open the mtx file and message file.                        */
/**************************************************************/

	if ((mtxfile = fopen(mtxFileName, "rb")) == NULL)
	{
		printf("Unable to open input file - %s\n", mtxFileName);
		exit(100);
	}

	if ((msgfile = fopen(msgFileName, "w")) == NULL)
	{
		fclose(mtxfile);
		printf("Unable to open message file - %s\n", msgFileName);
		exit(101);
	}

	// Write the name of the input file to the message file 
	writeFileName(mtxFileName, msgfile);

	// determine the length of the data file and read it into memory
	fseek(mtxfile, 0L, SEEK_END);
	dataLen = ftell(mtxfile);
	fseek(mtxfile, 0L, SEEK_SET);
	nameArea = malloc(dataLen);
	fread(nameArea, 1, dataLen, mtxfile);

	// Now, calculate the pointers to different parts of the file
	hdr = (META_FILE_HEADER *) nameArea;
	ptr = nameArea;
	tempPtr = nameArea + META_HDR_LEN;
	fvar = (VAR_FILE_REC *) tempPtr;
	nameArea += hdr->Names;
	UnameArea = (wchar_t*)(ptr + hdr->Unames);
	rvar = (REPEAT_REC *)(ptr + hdr->RepeatArea);;

	printReport(fvar, nameArea, UnameArea, hdr->Maxvar, msgfile);

	printRepeats(rvar, UnameArea, hdr->RepeatCount, msgfile);

	// print the header fields
	fprintf(msgfile, "\nHeader Length %d\n", hdr->HdrLength);
	fprintf(msgfile, "Version       %d\n", hdr->Version);
	fprintf(msgfile, "Max vars      %d\n", hdr->Maxvar);
	fprintf(msgfile, "Names offset  %d\n", hdr->Names);
	fprintf(msgfile, "Unicode ofs   %d\n", hdr->Unames);

	/* close the files */
	fclose(mtxfile);
	fclose(msgfile);

/**************************************************************/
/* Finished - return normally                                 */
/**************************************************************/

	return;
}

/*-------------*/
/* End Of File */
/*-------------*/
